/******************************************************************************
 *                                     __                                     *
 *                              <-----/@@\----->                              *
 *                             <-< <  \\//  > >->                             *
 *                               <-<-\ __ /->->                               *
 *                               Data /  \ Crow                               *
 *                                   ^    ^                                   *
 *                       (c) 2003 The Data Crow team                          *
 *                              info@datacrow.net                             *
 *                                                                            *
 *                                                                            *
 *       This library is free software; you can redistribute it and/or        *
 *        modify it under the terms of the GNU Lesser General Public          *
 *       License as published by the Free Software Foundation; either         *
 *     version 2.1 of the License, or (at your option) any later version.     *
 *                                                                            *
 *      This library is distributed in the hope that it will be useful,       *
 *      but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU       *
 *           Lesser General Public License for more details.                  *
 *                                                                            *
 *     You should have received a copy of the GNU Lesser General Public       *
 *    License along with this library; if not, write to the Free Software     *
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA   *
 *                                                                            *
 ******************************************************************************/

package net.sf.dc.console.components.lists;

import java.awt.Dimension;
import java.awt.Insets;
import java.awt.Point;
import java.awt.Rectangle;
import java.awt.event.ComponentEvent;
import java.awt.event.ComponentListener;
import java.util.ArrayList;
import java.util.Collection;
import java.util.Iterator;

import javax.swing.JList;
import javax.swing.SwingConstants;
import javax.swing.UIDefaults;
import javax.swing.UIManager;
import javax.swing.plaf.ListUI;

import net.sf.dc.console.ComponentFactory;
import net.sf.dc.console.components.lists.elements.DcAudioCDListHwElement;
import net.sf.dc.console.components.lists.elements.DcAudioTrackListElement;
import net.sf.dc.console.components.lists.elements.DcBookListHwElement;
import net.sf.dc.console.components.lists.elements.DcMediaObjectListElement;
import net.sf.dc.console.components.lists.elements.DcMovieListHwElement;
import net.sf.dc.console.components.lists.elements.DcMusicTrackListElement;
import net.sf.dc.console.components.lists.elements.DcObjectListElement;
import net.sf.dc.console.components.lists.elements.DcSimpleItemListElement;
import net.sf.dc.console.components.lists.elements.DcSoftwareListHwElement;
import net.sf.dc.console.components.lists.elements.DcTemplateListElement;
import net.sf.dc.console.components.renderers.DcObjectListRenderer;
import net.sf.dc.console.dialogs.MessageBox;
import net.sf.dc.core.modules.DcMediaModule;
import net.sf.dc.core.modules.DcModule;
import net.sf.dc.core.modules.DcModules;
import net.sf.dc.core.modules.contactperson.ContactPersonModule;
import net.sf.dc.core.objects.DcObject;
import net.sf.dc.core.objects.DcSimpleObject;
import net.sf.dc.core.objects.audiocd.AudioTrack;
import net.sf.dc.core.objects.musicalbum.MusicTrack;
import net.sf.dc.core.objects.template.ITemplate;
import net.sf.dc.resources.DcResources;

public class DcObjectList extends JList {

    private boolean hwElements;
    
    private static final String  uiClassID = "DcObjectListUI";
    private int columnsPerRow = 1;
    private int visibleColumnCount = 1;    
    
    static {
        UIDefaults ui = UIManager.getDefaults();
        if(!ui.contains(uiClassID))
            ui.put(uiClassID, "net.sf.dc.console.components.lists.DcObjectListUI");
    }
    
    public DcObjectList(  DcModule module, boolean hwElements, boolean wrap, boolean evenOddColors) {
        super(new DcObjectListModel());
        
        addComponentListener(new ComponentChangeListener());
        
        this.hwElements = hwElements;
        
        setBorder(ComponentFactory.getBorder());
        setCellRenderer(new DcObjectListRenderer(evenOddColors));
        setBackground(ComponentFactory.getDisabledColor());
        
        if ((module instanceof DcMediaModule|| module instanceof ContactPersonModule) && wrap)
            setLayoutOrientation(JList.HORIZONTAL_WRAP);
        else 
            setLayoutOrientation(JList.VERTICAL_WRAP);
        
        setMinimumSize(new Dimension(0,0));
    }    
    
    public String getUIClassID() {
        return uiClassID;
    }

    public int getVisibleColumnCount() {
        return visibleColumnCount;
    }
    
    public void setVisibleColumnCount(int visibleColumnCount) {
        int oldValue = this.visibleColumnCount;
        this.visibleColumnCount = Math.max(0, visibleColumnCount);
        firePropertyChange("visibleColumnCount", oldValue, visibleColumnCount);
    }
    
    public int getColumnsPerRow() {
        return columnsPerRow;
    }
    
    public void setColumnsPerRow(int columnsPerRow) {
        int oldValue = this.columnsPerRow;
        this.columnsPerRow = Math.max(1, columnsPerRow);
        if(this.columnsPerRow != oldValue)
            firePropertyChange("columnsPerRow", oldValue, columnsPerRow);
    }    

    public int locationToNearestIndex(Point location) {
        ListUI ui = getUI();
        if(ui == null || !(ui instanceof DcObjectListUI))
            return -1;

        return ((DcObjectListUI) ui).locationToNearestIndex(this, location);
    }
    
    public void setMaxVisibleColumnsPerRow() {
        Rectangle r = getVisibleRect();
        DcObjectListUI ui = (DcObjectListUI)getUI();
        int newMaxColPerRow = ui.getMaxColumnsPerRow(r.width);
        if(newMaxColPerRow != columnsPerRow)
            setColumnsPerRow(newMaxColPerRow);
    }    
    
    public int getLastVisibleIndex() {
        Rectangle r = getVisibleRect();
        Point visibleLR = new Point(0, (r.y + r.height) - 1);
        int index = locationToIndex(visibleLR);
        return index;
    }
    
    public int getFirstIndexOnLastVisibleRow() {
        Rectangle r = getVisibleRect();
        Point visibleLR = new Point(0, (r.y + r.height) - 1);
        return locationToIndex(visibleLR);
    }    
    
    private void checkScrollableParameters(Rectangle visibleRect, int orientation) {
        if (visibleRect == null)
            throw new IllegalArgumentException("visibleRect must be non-null");

        switch (orientation) {
            case SwingConstants.VERTICAL:
            case SwingConstants.HORIZONTAL:
                break;
            default:
                throw new IllegalArgumentException("orientation must be one of: VERTICAL, HORIZONTAL");
        }
    }    
    
    public Dimension getPreferredScrollableViewportSize() {
        Insets insets = getInsets();
        int dx = insets.left + insets.right;
        int dy = insets.top  + insets.bottom;

        int visibleRowCount    = getVisibleRowCount();
        int visibleColumnCount = getVisibleColumnCount();
        int fixedCellWidth     = getFixedCellWidth();
        int fixedCellHeight    = getFixedCellHeight();

        try {
            if ((fixedCellWidth > 0) && (fixedCellHeight > 0)) {
                int width  = (visibleColumnCount * fixedCellWidth)+ dx;
                int height = (visibleRowCount * fixedCellHeight) + dy;
                return new Dimension(width, height);
            } else if (getModel().getSize() > 0) {
                Rectangle r = getCellBounds(0, 0);
                int width  = (visibleColumnCount * r.width) + dx;
                int height = (visibleRowCount * r.height) + dy;
                return new Dimension(width, height);
            } else {
                fixedCellWidth = (fixedCellWidth > 0) ? fixedCellWidth : 256;
                fixedCellHeight = (fixedCellHeight > 0) ? fixedCellHeight : 16;
                return new Dimension(fixedCellWidth, fixedCellHeight * visibleRowCount);
            }
        } catch (Exception exp) {
            return new Dimension(fixedCellWidth, fixedCellHeight * visibleRowCount);
        }
    } 

    public int getScrollableUnitIncrement(Rectangle visibleRect, int orientation, int direction) {
        checkScrollableParameters(visibleRect, orientation);

        int firstIndex = getFirstVisibleIndex();
        if (getLayoutOrientation() == JList.VERTICAL_WRAP)
            return 40;
        
        int increment;
        int row    = firstIndex / columnsPerRow;
        int column = firstIndex % columnsPerRow;

        if (orientation == SwingConstants.HORIZONTAL) {

            if (direction > 0) {
                Rectangle r = getCellBounds(firstIndex, firstIndex);
                increment = (r == null) ? 0 : r.width - (visibleRect.x - r.x);
            } else {
                Rectangle r = getCellBounds(firstIndex, firstIndex);

                if ((r.x == visibleRect.x) && (column == 0)) {
                    increment = 0;
                } else if (r.x == visibleRect.x) {
                   Rectangle prevR = getCellBounds(firstIndex - 1, firstIndex - 1);
                   increment = (prevR == null) ? 0 : prevR.width;
                } else {
                    increment = visibleRect.x - r.x;
                }
            }
        } else {
            if (direction > 0) {
                Rectangle r = getCellBounds(firstIndex, firstIndex);
                increment = (r == null) ? 0 : r.height - (visibleRect.y - r.y);
            } else {
                Rectangle r = getCellBounds(firstIndex, firstIndex);
                if ((r.y == visibleRect.y) && (row == 0))  {
                    increment = 0;
                } else if (r.y == visibleRect.y) {
                    Rectangle prevR = getCellBounds(firstIndex - 1, firstIndex - 1);
                    increment = (prevR== null) ? 0 : prevR.height;
                } else {
                    increment = visibleRect.y - r.y;
                }
            }
        }
        return increment;
    }
    
    public void setSelectionMode(int selectionMode) {
        getSelectionModel().setSelectionMode(selectionMode);        
    }
    
    public void setSelected(DcObject dco, int field) {
        for (int row = 0 ; row < getDcModel().getSize(); row++) {
            DcObjectListElement element = (DcObjectListElement) getDcModel().getElementAt(row);
            if (element.getDcObject() != null) {
                if (dco.getDisplayString(field).equals(element.getDcObject().getDisplayString(field))) {
                    setSelectedIndex(row);
                    break;
                }
            }
        }
    }
    
    public void setSelected(Collection dcos) {
        int[] indices = new int[dcos.size()];
        
        int counter = 0;
        for (Iterator iter = dcos.iterator(); iter.hasNext(); counter++) {
            Object o = iter.next();
            for (int row = 0 ; row < getDcModel().getSize(); row++) {
                DcObjectListElement element = (DcObjectListElement) getDcModel().getElementAt(row);
                if (element.getDcObject() != null) {
                    if (element.getDcObject().equals(o)) {
                        indices[counter] = row;
                        break;
                    }
                }
            }
        }
        setSelectedIndices(indices);
    }
    
    public DcObjectListModel getDcModel() {
        return (DcObjectListModel) getModel();
    }
    
    public int[] getChangedIndices() {
        Collection rows = new ArrayList();
        for (int i = 0 ; i < getDcModel().getSize(); i++) {
            DcObjectListElement element = (DcObjectListElement) getDcModel().getElementAt(i);
            if (element.getDcObject() != null && element.getDcObject().isChanged()) 
                rows.add(new Integer(i));
        }
        
        int[] indices = new int[rows.size()];
        int counter = 0;
        for (Iterator iter = rows.iterator(); iter.hasNext(); counter++) {
            indices[counter] = ((Integer) iter.next()).intValue();
        }
        return indices;        
    }
    
    public void markAsUnchanged() {
        int[] rows = getChangedIndices();
        for (int i = 0; i < rows.length; i++) {
            DcObjectListElement element = (DcObjectListElement) getDcModel().getElementAt(i);
            element.getDcObject().markAsUnchanged(); 
        }
    }    
    
    public DcObject getObjectForID(String ID) {
        DcObjectListElement element = getElementForID(ID);
        return element != null ? element.getDcObject() : null;
    }    

    public DcObjectListElement getElementForID(String ID) {
        for (int i = 0 ; i < getDcModel().getSize(); i++) {
            DcObjectListElement element = (DcObjectListElement) getDcModel().getElementAt(i);
            if (element.getDcObject() != null && element.getDcObject().getID().equals(ID)) 
                return element;
        }
        return null;
    }      
    
    public Collection getObjects() {
        Collection objects = new ArrayList();
        for (int i = 0 ; i < getDcModel().getSize(); i++) {
            DcObjectListElement element = (DcObjectListElement) getDcModel().getElementAt(i);
            if (element.getDcObject() != null) 
                objects.add(element.getDcObject());
        }
        return objects;
    }
    
    public Collection getSelectedObjects() {
        int[] indices = getSelectedIndices();
        Collection objects = new ArrayList();

        if (indices == null || indices.length == 0) {
            new MessageBox(DcResources.getText("msgNoRowsSelected"), MessageBox._WARNING);
        } else {
            for (int i = 0; i < indices.length; i++) {
                DcObjectListElement element = (DcObjectListElement) getDcModel().getElementAt(indices[i]);
                DcObject dco = element.getDcObject();
                dco.setRow(indices[i]);
                objects.add(dco);
            }
        }
        return objects;
    }
    
    public void update() {
        for (int i = 0 ; i < getDcModel().getSize(); i++) {
            DcObjectListElement element = (DcObjectListElement) getDcModel().getElementAt(i);
            element.update();
        }
        repaint();
    }
    
    public void remove() {
        getDcModel().remove(getSelectedIndex());
    }

    public void remove(int[] indices) {
        for (int i = 0; i < indices.length; i++) {
            getDcModel().removeElementAt(indices[i]);
        }
    }    
    
    public void remove(String[] ids) {
        for (int i = 0; i < ids.length; i++) {
            getDcModel().removeElement(getElementForObjectID(ids[i]));
        }
    }

    public void updateElement(int index, DcObject dco, boolean overwrite, boolean allowDeletes, boolean mark) {
        DcObjectListElement element = (DcObjectListElement) getDcModel().getElementAt(index);
        element.update(dco, overwrite, allowDeletes, mark);
        
        try {
            setSelectedValue(getDcModel().getElementAt(index), true);
        } catch (Exception ignore) {}
    }
    
    public void updateElement(String objectID, DcObject dco, boolean overwrite, boolean allowDeletes, boolean mark) {
        DcObjectListElement element = getElementForObjectID(objectID);
        if (element != null)
            element.update(dco, overwrite, allowDeletes, mark);
        
        try {
            setSelectedValue(getElementForObjectID(objectID), true);
        } catch (Exception ignore) {}
    }
    
    public void clear() {
        try {
           for (int i = 0 ; i < getDcModel().getSize(); i++) {
              DcObjectListElement element = (DcObjectListElement) getDcModel().getElementAt(i);
              element.clear();
              remove(element);
           }
        } catch (Exception ignore) {}
        
        getDcModel().clear();
        getDcModel().removeAllElements();
        
        repaint();
    }
    
    private DcObjectListElement getElementForObjectID(String objectID) {
        for (int i = 0 ; i < getDcModel().getSize(); i++) {
            DcObjectListElement element = (DcObjectListElement) getDcModel().getElementAt(i);
            if (objectID.equals(element.getDcObject().getID()))
                return element;
        }
        return null;
    }    
    
    public DcObject getSelectedObject() {
        DcObjectListElement element = (DcObjectListElement) getSelectedValue();
        
        DcObject dco = null;
        if (element != null) 
            dco = element.getDcObject(); 
        
        return dco;
    }
    
    public void listenersEnabled(boolean b) {
        getDcModel().setListenersEnabled(b);        
    }
    
    public void add(DcObject dco) {
        if (dco.getID() == null || dco.getID().equals("")) 
            dco.setIDs();        
        
        dco.markAsUnchanged();
        getDcModel().addElement(getDisplayElement(dco));
        
        try {
            ensureIndexIsVisible(getModel().getSize());
        } catch (Exception exp) {}

    }
    
    public void add(DcObject[] objects) {
        listenersEnabled(false);
        
        for (int i = 0; i < objects.length; i++) {
            DcObject dco = objects[i];
            if (dco.getID() == null || dco.getID().equals("")) 
                dco.setIDs(); 
            
            add(dco);
        }
        
        listenersEnabled(true);
        
        getDcModel().fireIntervalAdded(getDcModel(), 0, objects.length - 1);
    }    
    
    public void add(Collection objects) {
        listenersEnabled(false);
        
        for (Iterator iter = objects.iterator(); iter.hasNext(); ) {
            DcObject dco = (DcObject) iter.next();
            
            if (dco.getID() == null || dco.getID().equals("")) 
                dco.setIDs(); 
            
            add(dco);
        }
        
        listenersEnabled(true);
        
        getDcModel().fireIntervalAdded(getDcModel(), 0, objects.size() - 1);
    }
    
    public DcObjectListElement getDisplayElement(DcObject dco) {
        DcObjectListElement element = null;
        DcModule module = dco.getModule();
        
        if (hwElements) {
            if (module.getIndex() == DcModules._AUDIOCD) 
                element = new DcAudioCDListHwElement(dco);
            else if (module.getIndex() == DcModules._SOFTWARE)
                element = new DcSoftwareListHwElement(dco);
            else if (module.getIndex() == DcModules._MOVIE)
                element = new DcMovieListHwElement(dco);
            else if (module.getIndex() == DcModules._BOOK)
                element = new DcBookListHwElement(dco);
        } else {
            if (dco instanceof AudioTrack)
                element = new DcAudioTrackListElement(dco);
            else if (dco instanceof ITemplate)
                element = new DcTemplateListElement(dco);
            else if (dco instanceof MusicTrack)
                element = new DcMusicTrackListElement(dco);
            else if (dco instanceof DcSimpleObject)
                element = new DcSimpleItemListElement(dco);
            else 
                element = new DcMediaObjectListElement(dco);       
        }

        return element;
    }   
    
    private class ComponentChangeListener implements ComponentListener {

        public void componentHidden(ComponentEvent ce) {}

        public void componentMoved(ComponentEvent ce) {}

        public void componentResized(ComponentEvent ce) {
            setMaxVisibleColumnsPerRow();
        }

        public void componentShown(ComponentEvent ce) {}
    }    
}
