/******************************************************************************
 *                                     __                                     *
 *                              <-----/@@\----->                              *
 *                             <-< <  \\//  > >->                             *
 *                               <-<-\ __ /->->                               *
 *                               Data /  \ Crow                               *
 *                                   ^    ^                                   *
 *                       (c) 2003 The Data Crow team                          *
 *                              info@datacrow.net                             *
 *                                                                            *
 *                                                                            *
 *       This library is free software; you can redistribute it and/or        *
 *        modify it under the terms of the GNU Lesser General Public          *
 *       License as published by the Free Software Foundation; either         *
 *     version 2.1 of the License, or (at your option) any later version.     *
 *                                                                            *
 *      This library is distributed in the hope that it will be useful,       *
 *      but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU       *
 *           Lesser General Public License for more details.                  *
 *                                                                            *
 *     You should have received a copy of the GNU Lesser General Public       *
 *    License along with this library; if not, write to the Free Software     *
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA   *
 *                                                                            *
 ******************************************************************************/

package net.sf.dc.console.components.panels;

import java.awt.Color;
import java.awt.Component;
import java.awt.Dimension;
import java.awt.Font;
import java.awt.GridBagConstraints;
import java.awt.Insets;
import java.awt.event.MouseEvent;
import java.awt.event.MouseListener;
import java.io.StringReader;
import java.util.Iterator;
import java.util.LinkedList;

import javax.swing.JMenuItem;
import javax.swing.JPanel;
import javax.swing.JPopupMenu;
import javax.swing.JScrollPane;
import javax.swing.JTabbedPane;
import javax.swing.JTextPane;
import javax.swing.SwingUtilities;
import javax.swing.event.ChangeEvent;
import javax.swing.event.ChangeListener;
import javax.swing.text.html.HTMLDocument;
import javax.swing.text.html.HTMLEditorKit;

import net.sf.dc.console.ComponentFactory;
import net.sf.dc.console.Layout;
import net.sf.dc.console.components.DcPictureField;
import net.sf.dc.console.forms.ItemForm;
import net.sf.dc.console.views.MasterView;
import net.sf.dc.core.IconLibrary;
import net.sf.dc.core.Repository;
import net.sf.dc.core.actions.ToggleQuickViewAction;
import net.sf.dc.core.modules.DcModule;
import net.sf.dc.core.modules.DcModules;
import net.sf.dc.core.objects.DcObject;
import net.sf.dc.core.objects.Picture;
import net.sf.dc.messages.Messages;
import net.sf.dc.resources.DcResources;
import net.sf.dc.settings.DcSettings;
import net.sf.dc.settings.definitions.DcFieldDefinition;
import net.sf.dc.settings.definitions.DcFieldDefinitions;
import net.sf.ext.util.Utilities;

public abstract class QuickViewPanel extends JPanel {
    
    public static final int _DIRECTION_HORIZONTAL = 0;
    public static final int _DIRECTION_VERTICAL = 1;    
    
    protected final DcModule module;
    protected final boolean showImages;
    
    protected DcObject dco;
    private LinkedList pictures = new LinkedList();
    private LinkedList imagePanels = new LinkedList();

    private JTextPane descriptionPane;
    private JScrollPane scroller;
    private final JTabbedPane tabbedPane = ComponentFactory.getTabbedPane();
    
    private final HTMLEditorKit kit = new HTMLEditorKit();
    private final HTMLDocument document = new HTMLDocument();
    
    public QuickViewPanel(DcModule module, boolean showImages) {
        this.module = module;
        this.showImages = showImages;
        setLayout(Layout.getGBL());
        tabbedPane.addChangeListener(new TabListener());
        buildPanel();
    }    
    
    public void setFont(Font font) {
        super.setFont(font);

        if (tabbedPane != null)
            tabbedPane.setFont(ComponentFactory.getSystemFont());
        
        if (dco != null) {
            setObject(dco);            
        }
    }
    
    private synchronized void loadImage() {
        int index = tabbedPane.getSelectedIndex() - 1;
        Picture picture =  (Picture) pictures.get(index);
        picture.loadImage();
        
        JPanel panel =  (JPanel) imagePanels.get(index);
        Component[] components =  panel.getComponents();
        for (int i = 0; i < components.length; i++) {
            if (components[i] instanceof DcPictureField) {
                DcPictureField picField = (DcPictureField) components[i];
                picField.setValue(picture.getValue(Picture._D_IMAGE));
                picField.repaint();
            }
        }
    }
    
    public synchronized void setImages() {
        try {
            clearImages();
    
            if (dco == null)
                return;
            
            DcFieldDefinitions definitions = dco.getModule().getFieldDefinitions(); 
            
            for (Iterator iter =  definitions.iterator(); iter.hasNext(); ) {
                DcFieldDefinition definition = (DcFieldDefinition) iter.next();
                if (    dco.isEnabled(definition.getIndex()) && 
                        dco.getField(definition.getIndex()).getValueType() == Repository.datatypes._PICTURE) {
                    
                    Picture picture = (Picture) dco.getValue(definition.getIndex());
                    if (picture != null) {
                        pictures.add(picture);    
                        
                        DcPictureField picField = ComponentFactory.getPictureField(true, false, false);
                        picField.setBorder(ComponentFactory.getBorder());
                        
                        JPanel panel = new JPanel();
                        panel.setLayout(Layout.getGBL());
                        panel.addMouseListener(new QuickViewMouseListener());
                        panel.add(picField, Layout.getGBC(0, 0, 1, 1, 10.0, 10.0,
                                GridBagConstraints.NORTHWEST, GridBagConstraints.BOTH,
                                new Insets(2, 2, 2, 2), 0, 0));
                            
                        tabbedPane.addTab(dco.getLabel(definition.getIndex()), IconLibrary._icoPicture, panel);
                        imagePanels.add(panel);
                    }
                }
            }
        } catch (Exception exp) {
            Messages.add(exp);
        }
    }
    
    public synchronized void setObject(DcObject dco) {
        
        try {
            if (dco == null)
                return;
            
            if (DcModules.getCurrentModule().getIndex() == DcModules._GENERIC &&
                DcModules.getCurrentModule().getCurrentSearchView().getIndex() == MasterView._TABLE_VIEW)
                dco.reload();
            
            int tab = tabbedPane.getSelectedIndex();
            clear();
            this.dco = dco;
            
            String html = "<html><body " + getBodyStyle() + ">\n" +
                            getDescriptionTable() +
                          "</body> </html>";
            
            setDescriptionHtml(html);
            
            try {
                descriptionPane.setCaretPosition(0);
            } catch (Exception exp) {}
            
            setImages();
            
            boolean error = true;
            tab += 1;
            
            // prevent endless loop
            int counter = 0;
            while (error && counter < 6) {
                counter ++;
                try {
                    tab -= 1;
                    tabbedPane.setSelectedIndex(tab);
                    error = false;
                } catch (Exception ignore) {}
            }
        } catch (Exception exp) {
            Messages.add(exp);
        }
    }
    
    public synchronized void clear() {
        Dimension dim = getSize();
        setPreferredSize(dim);
        if (dco != null) {
            dco = null;
            descriptionPane.setText("");

            try {
                tabbedPane.setSelectedIndex(0);
            } catch (Exception exp) {}
            
            clearImages();
        }
    }
    
    private synchronized void clearImages() {
        tabbedPane.removeAll();
        tabbedPane.addTab("Description", IconLibrary._icoInformation ,scroller);
        
        for (Iterator iter = imagePanels.iterator(); iter.hasNext(); ) {
            JPanel panel = (JPanel) iter.next();
            Component[] components = panel.getComponents();
            for (int i = 0; i < components.length; i++) {
                try {
                    if (components[i] instanceof DcPictureField)
                        ((DcPictureField) components[i]).flush();
                } catch (Exception ignore) {}
            }
            panel.removeAll();
        }
        pictures.clear();
        imagePanels.clear();
    }
    
    protected abstract String getDescriptionTable();
    
    protected synchronized void setDescriptionHtml(String html) {
        try {
            descriptionPane.read(new StringReader(html), "Data Crow");
        } catch (Exception exp) {
            Messages.add(exp);
        }
    }
    
    protected String addTableRow(String table, int index, int textDirection) {
        if (dco.isEnabled(index)) {
            Font font = DcSettings.getValueAsFont(Repository.settings.stSystemFontNormal);
            boolean horizontal = textDirection == _DIRECTION_HORIZONTAL;

            String value = dco.getDisplayString(index);
            String realValue = dco.getValue(index) == null ? "" : dco.getValue(index).toString();
            String displayString = value == null ? "" : value.toString();
            if (realValue != null && displayString.length() > 0 ) {
                table += "<tr><td>\n";
                table += "<b>" + dco.getLabel(index) + "</b>";
                
                if (!horizontal) {
                    table += "</td></tr>";
                    table += "<tr><td>\n";
                } else {
                    table += " ";
                }
                
                if (font.getStyle() == Font.BOLD) 
                    value = "<b>" + value + "</b>";
                else
                    value += "<br><br>";
                    
                table += value;
                table += "</td></tr>";
            } else if (dco.getField(index).getValueType() == Repository.datatypes._PICTURE) {
                Picture picture = (Picture) dco.getValue(index);
                pictures.add(picture);
            }
        }
        return table;
    }    
    
    private synchronized void buildPanel() {
        // description panel
        descriptionPane = ComponentFactory.getTextPane();
        descriptionPane.setEditorKit(kit);
        descriptionPane.setDocument(document);
        descriptionPane.setEditable(false);
        descriptionPane.setBounds(1,1,1,1);
        descriptionPane.addMouseListener(new QuickViewMouseListener());

        scroller = new JScrollPane(descriptionPane);
        scroller.setVerticalScrollBarPolicy(JScrollPane.VERTICAL_SCROLLBAR_ALWAYS);
        scroller.setHorizontalScrollBarPolicy(JScrollPane.HORIZONTAL_SCROLLBAR_NEVER);
        
        // tabbed pane
        tabbedPane.addTab("Description", IconLibrary._icoInformation ,scroller);
        add(tabbedPane, Layout.getGBC(0, 0, 1, 1, 1.0, 1.0,
            GridBagConstraints.NORTHWEST, GridBagConstraints.BOTH,
            new Insets(4, 0, 3, 0), 0, 0));
        
        String html = "<html><body " + getBodyStyle() + "></body> </html>";
        setDescriptionHtml(html);
    }    
    
    private void showPopupMenu(int x, int y) {
        PopupMenu popupMenu = new PopupMenu();
        popupMenu.show(this, x, y);
    }    
    
    protected String getBodyStyle() {
        Color colorEven = DcSettings.getValueAsColor(Repository.settings.stEvenRowColor);
        Color colorOdd = DcSettings.getValueAsColor(Repository.settings.stOddRowColor);
        
        
        Color color = colorEven.getRGB() < colorOdd.getRGB() ? colorOdd : colorEven;
        String background = Utilities.getHexColor(color);
       
        descriptionPane.setBackground(color);
        for (Iterator iter = imagePanels.iterator(); iter.hasNext(); ) {
            ((JPanel) iter.next()).setBackground(color);
        }
        
        String hexColor = Utilities.getHexColor(ComponentFactory.getCurrentForegroundColor());
        Font font = DcSettings.getValueAsFont(Repository.settings.stSystemFontNormal);
        
        String fontstyle = "";
        if (font.isItalic())
            fontstyle = ";font-style:italic";
        else if (font.isBold())
            fontstyle = ";font-weigth:bolder";
        
        String style = "style=" + '"' + 
                       ";background:" + background + 
                       ";font-family:" + font.getFamily() + 
                       ";font-size:" + font.getSize() + fontstyle +
                       ";color:" + hexColor + ";" + '"';
        return style;
    }
    
    public class PopupMenu extends JPopupMenu {
        public PopupMenu() {
            JMenuItem menuToggleLabels = ComponentFactory.getMenuItem(DcResources.getText("lblToggleQuickView"));
            menuToggleLabels.addActionListener(new ToggleQuickViewAction());
            this.add(menuToggleLabels);
        }
    }   
    
    private class TabListener implements ChangeListener {
        public void stateChanged(ChangeEvent evt) {
            JTabbedPane pane = (JTabbedPane) evt.getSource();
            
            if (pane.getSelectedIndex() > 0)
                loadImage();
        }
    } 
    
    private class QuickViewMouseListener implements MouseListener {
        public void mouseReleased(MouseEvent e) {
            if (SwingUtilities.isRightMouseButton(e)) {
                showPopupMenu(e.getX(), e.getY());
            } 

            if (e.getClickCount() == 2 && dco != null) {
            	dco.markAsUnchanged();
                ItemForm itemform = new ItemForm(null, false, true, dco, module.isImdbUpdateAllowed(), module.isAmazonUpdateAllowed(), false);
                itemform.setVisible(true);
            }
        }
        public void mouseEntered(MouseEvent e) {}
        public void mouseExited(MouseEvent e) {}
        public void mousePressed(MouseEvent e) {}
        public void mouseClicked(MouseEvent e) {}
    }
}
