/******************************************************************************
 *                                     __                                     *
 *                              <-----/@@\----->                              *
 *                             <-< <  \\//  > >->                             *
 *                               <-<-\ __ /->->                               *
 *                               Data /  \ Crow                               *
 *                                   ^    ^                                   *
 *                       (c) 2003 The Data Crow team                          *
 *                              info@datacrow.net                             *
 *                                                                            *
 *                                                                            *
 *       This library is free software; you can redistribute it and/or        *
 *        modify it under the terms of the GNU Lesser General Public          *
 *       License as published by the Free Software Foundation; either         *
 *     version 2.1 of the License, or (at your option) any later version.     *
 *                                                                            *
 *      This library is distributed in the hope that it will be useful,       *
 *      but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU       *
 *           Lesser General Public License for more details.                  *
 *                                                                            *
 *     You should have received a copy of the GNU Lesser General Public       *
 *    License along with this library; if not, write to the Free Software     *
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA   *
 *                                                                            *
 ******************************************************************************/

package net.sf.dc.console.components.panels;

import java.awt.Font;
import java.awt.GridBagConstraints;
import java.awt.Insets;
import java.util.ArrayList;
import java.util.Collection;
import java.util.Collections;
import java.util.Comparator;
import java.util.Iterator;
import java.util.List;

import javax.swing.JPanel;
import javax.swing.JScrollPane;
import javax.swing.event.TreeSelectionEvent;
import javax.swing.event.TreeSelectionListener;
import javax.swing.tree.DefaultMutableTreeNode;
import javax.swing.tree.DefaultTreeModel;
import javax.swing.tree.TreePath;

import net.sf.dc.console.ComponentFactory;
import net.sf.dc.console.Layout;
import net.sf.dc.console.components.DcTree;
import net.sf.dc.console.menu.DcTreePanelMenuBar;
import net.sf.dc.console.views.View;
import net.sf.dc.core.Repository;
import net.sf.dc.core.data.DataFilter;
import net.sf.dc.core.data.DataFilters;
import net.sf.dc.core.data.DataManager;
import net.sf.dc.core.modules.DcMappingModule;
import net.sf.dc.core.modules.DcModule;
import net.sf.dc.core.modules.DcModules;
import net.sf.dc.core.objects.DcField;
import net.sf.dc.core.objects.DcMapping;
import net.sf.dc.core.objects.DcObject;
import net.sf.dc.messages.Messages;
import net.sf.dc.resources.DcResources;
import net.sf.dc.settings.DcSettings;

public class TreePanel extends JPanel {

    public static final int  _OBJECT_ADDED = 0;
    public static final int  _OBJECT_UPDATED = 1;
    public static final int  _OBJECT_REMOVED = 2;
    
    private final int module;
    private final View view;
    private int[] fields;
    
    private DcTree tree;
    private JScrollPane scroller;
    private DefaultMutableTreeNode top;
    
    private Object currentUserObject;
    private DcTreePanelMenuBar menu;
    
    private Collection objects = new ArrayList();
    
    private final TreeSelectionListener tsl = new DcTreeSelectionListener();
    
    public TreePanel(int module, View view) {
        this.view = view;
        this.menu = new DcTreePanelMenuBar(module, this);
        this.module = module;
        setLayout(Layout.getGBL());
        initializeMenu();
        fields = DcSettings.getValueAsIntegerArray(DcModules.get(module).getGroupedBySettingsKey());
    }
    
    public Object getLastSelectedPathComponent() {
        return tree.getLastSelectedPathComponent();
    }
    
    public synchronized void groupBy(int[] fields) {
        this.fields = fields;
        DcSettings.setValue(DcModules.get(module).getGroupedBySettingsKey(), fields);
        
        Collection c = getObjects();
        
        DcObject[] objects = new DcObject[c.size()];
        objects = (DcObject[]) c.toArray(objects);
        
        buildTree(objects);
    }

    public void setFont(Font font) {
        if (tree != null) {
            tree.setFont(font);
            menu.setFont(font);
        }
    }
    
    public void add(DcObject[] objects) {
        fields = DcSettings.getValueAsIntegerArray(DcModules.get(module).getGroupedBySettingsKey());
        buildTree(objects);
    }

    public synchronized void add(DcObject dco) {
        if (tree == null)  {
            initializeTree();
            buildTree(objects);
        }
        
        if (tree != null) {
            if (!objects.contains(dco)) {
                objects.add(dco);
                revalidateTree(dco, _OBJECT_ADDED);
            }
        }
    }

    public synchronized void remove(DcObject dco) {
        if (tree != null) {
            tree.removeTreeSelectionListener(tsl);
            if (objects.contains(dco)) {
                objects.remove(dco);
                revalidateTree(dco, _OBJECT_REMOVED);
            }
            tree.addTreeSelectionListener(tsl);
        }
    }
        
    public synchronized void remove(String s) {
        DcObject dco = DcModules.get(module).getDcObject();
        dco.setValue(DcObject._ID, s);
        remove(dco);
    }     
    
    public synchronized void remove(String[] ids) {
        for (int i = 0; i < ids.length; i++) 
            remove(ids[i]);
    } 

    
    public synchronized void update(DcObject dco) {
        if (tree != null) {
            int pos = ((List) objects).indexOf(dco);
            if (pos != -1) {
                DcObject o = (DcObject) ((List) objects).get(pos);
                o.reload();
                revalidateTree(o, _OBJECT_UPDATED);
            }
        }
    }    
    
    private void removeElement(DcObject dco) {
        
        Collection nodes = fields != null && fields.length > 0 ? findNodes(dco, top) : new ArrayList();
        
        for (Iterator iter = nodes.iterator(); iter.hasNext(); ) {
            DefaultMutableTreeNode parentNode = (DefaultMutableTreeNode) iter.next();
            NodeElement parentElement = (NodeElement) parentNode.getUserObject();
            parentElement.removeValue(dco);
            if (fields.length > 1) {
                Collection childNodes = findNodes(dco, parentNode);
                for (Iterator iterChildren = childNodes.iterator(); iterChildren.hasNext(); ) {
                    DefaultMutableTreeNode childNode = (DefaultMutableTreeNode) iterChildren.next();
                    NodeElement childElement = (NodeElement) childNode.getUserObject();
                    childElement.removeValue(dco);
                }
            }
        }
        
        removeEmptyLeafs(top);
    }
    
    private boolean clearSafely = true;

    private void addElement(DcObject dco, boolean select) {
        addElement(dco, top, 0, select);
    }
    
    private synchronized void addElement(DcObject dco, DefaultMutableTreeNode parent, int level, boolean select) {
        try {
            
            boolean grouped = fields != null;
            for (int i = 0; grouped && i < fields.length; i++)
                grouped = fields[i] != -1;
            
            if (!grouped)
                return;
            
            Collection nodes = fields.length > 0 ? findNodes(dco, fields[level], parent) : new ArrayList();
            clearSafely = false;
            
            if (level == 0 && fields != null && fields.length > 0 && nodes.size() == 0) {
//              no nodes available, create all
                DefaultMutableTreeNode parentNode = parent;
                for (int i = 0; i < fields.length; i++) {
                    parentNode = createLeaf(parentNode, dco, i);
                }
                
                if (select) 
                    setSelected(parentNode);
                
            } else if (nodes.size() > 0) {
//              add and create where needed
                for (Iterator iter = nodes.iterator(); iter.hasNext(); ) {
                    DefaultMutableTreeNode parentNode = (DefaultMutableTreeNode) iter.next();
                    NodeElement parentElement = (NodeElement) parentNode.getUserObject();
                    parentElement.addValue(dco);
                    if (fields.length > 1 && level < 1) 
                        addElement(dco, parentNode, 1, select);
                    else if (select) 
                        setSelected(parentNode);
                }
            } else if (isActive()) {
                DefaultMutableTreeNode node = createLeaf(parent, dco, level);
                if (select) setSelected(node);
            } else {
                DefaultMutableTreeNode node = (DefaultMutableTreeNode) parent.getChildAt(0);
                ((NodeElement) node.getUserObject()).addValue(dco);
                if (select) setSelected(node);
            }
        } catch (Exception e) {
            Messages.add(e);
        } finally {
            clearSafely = true;
        }
    }
    
    private void setSelected(DefaultMutableTreeNode node) {
        DefaultTreeModel model = (DefaultTreeModel) tree.getModel();
        TreePath treePath = new TreePath(model.getPathToRoot(node));
        tree.setSelectionPath(treePath);
        tree.scrollPathToVisible(treePath);
    }
    
    private synchronized DefaultMutableTreeNode createLeaf(DefaultMutableTreeNode parent, DcObject dco, int level) {
        int fldIdx = fields[level]; 
        
        DefaultMutableTreeNode node = null;
        if (DcModules.get(module).getField(fldIdx).getValueType() == Repository.datatypes._DCOBJECTCOLLECTION) {
            Collection mappings = (Collection) dco.getValue(fldIdx);
            if (mappings != null) {
                for (Iterator iter = mappings.iterator(); iter.hasNext(); ) {
                    DcObject so = ((DcMapping) iter.next()).getSimpleObject();
                    NodeElement ne = new NodeElement();
                    ne.setKey(so.toString());
                    ne.addValue(dco);
                    node = new DefaultMutableTreeNode(ne);
                    parent.add(node);
                    insertNode(node, parent);
                }
            }
        } else {
            NodeElement element = new NodeElement();
            element.addValue(dco);
            element.setKey(dco.getDisplayString(fields[level]));

            node = new DefaultMutableTreeNode(element);
            parent.add(node);
            insertNode(node, parent);
        }
        
        return node;
    }
    
    /**
     * Inserts a node. The node needs to have a valid user object defined
     * to insert it at the right position (ordering)
     */
    private synchronized void insertNode(DefaultMutableTreeNode node, DefaultMutableTreeNode parent) {
        DefaultTreeModel model = (DefaultTreeModel) tree.getModel();
        
        List elements = new ArrayList();
        for (int i = 0; i < parent.getChildCount(); i++) {
            elements.add(((DefaultMutableTreeNode) parent.getChildAt(i)).getUserObject().toString());
        }
        
        elements.add(node.getUserObject().toString());
        Collections.sort(elements);
        int idx = elements.indexOf(node.getUserObject().toString());
        
        model.insertNodeInto(node, parent, idx);
    }
    
    private synchronized void removeEmptyLeafs(DefaultMutableTreeNode parent) {
        try {
            int count = parent.getChildCount();
            
            tree.removeTreeSelectionListener(tsl);
            for (int i = count; i > 0; i--) {
                DefaultMutableTreeNode parentNode = (DefaultMutableTreeNode) parent.getChildAt(i - 1);
                
                if (parentNode.getUserObject() instanceof NodeElement) {
                    NodeElement ne = (NodeElement) parentNode.getUserObject();
                    if (ne.getValues().size() == 0) {
                        tree.setSelectionInterval(0, 0);
                        DefaultTreeModel model = (DefaultTreeModel) tree.getModel();
                        model.removeNodeFromParent(parentNode);
                    } else {
                        removeEmptyLeafs(parentNode);
                    }
                }
            }
        } catch (Exception exp) {
            Messages.add(exp);
        } finally {
            tree.addTreeSelectionListener(tsl);
        }
    }
    
    private synchronized void revalidateTree(DcObject dco, int modus) {
        if (modus == _OBJECT_REMOVED) {
            removeElement(dco);
        } else if (modus == _OBJECT_UPDATED) {
            removeElement(dco);
            addElement(dco, (dco.isBatch() ? dco.isEndOfBatch() : true));
        } else if (modus == _OBJECT_ADDED) {
            addElement(dco, true);
        }

        repaint();
        revalidate();
    }
    
    private synchronized Collection findNodes(DcObject dco, int field, DefaultMutableTreeNode parentNode) {
        Collection nodes = new ArrayList();
        if (field == -1)
            return nodes;
            
        if (dco.getField(field).getValueType() == Repository.datatypes._DCOBJECTCOLLECTION) {
            
            Collection c = (Collection) dco.getValue(field);
            for (Iterator iter = c.iterator(); iter.hasNext(); ) {
                DcMapping mapping = (DcMapping) iter.next();
                String ID = (String) mapping.getValue(DcMapping._B_SIMPLE_OBJECT_ID);
                DcObject so = DataManager.getObject(mapping.getSimpleModuleIdx(), ID);
                DefaultMutableTreeNode node = findNode(so.toString(), parentNode);
                if (node != null) { 
                    nodes.add(node);
                } else {
                    
                    // workaround
                    NodeElement ne = new NodeElement();
                    ne.setKey(so.toString());
                    ne.addValue(dco);
                    node = new DefaultMutableTreeNode(ne);
                    parentNode.add(node);
                    insertNode(node, parentNode);
                }
            }
        } else {
            DefaultMutableTreeNode node = findNode(dco.getDisplayString(field), parentNode);
            if (node != null) nodes.add(node);
        }
        
        return nodes;
    }

    private synchronized DefaultMutableTreeNode findNode(String key, DefaultMutableTreeNode parentNode) {
        int count = parentNode != null ? parentNode.getChildCount() : 0;
        for (int i = 0; i < count; i++) {
            DefaultMutableTreeNode node = (DefaultMutableTreeNode) parentNode.getChildAt(i);
            
            if (node.getUserObject() instanceof NodeElement) {
                NodeElement ne = (NodeElement) node.getUserObject();
                if (ne.getKey().equals(key))
                    return node;
            }
        }
        return null;
    }    
    
    private synchronized Collection findNodes(DcObject dco, DefaultMutableTreeNode parentNode) {
        Collection nodes = new ArrayList();
        
        int count = parentNode.getChildCount();
        for (int i = 0; i < count; i++) {
            DefaultMutableTreeNode node = (DefaultMutableTreeNode) parentNode.getChildAt(i);
            
            if (node.getUserObject() instanceof NodeElement) {
                NodeElement ne = (NodeElement) node.getUserObject();
                if (ne.getValues().contains(dco))
                    nodes.add(node);
            }
        }
        return nodes;
    }
    
    public void collapseAll() {
        if (top == null) return;
        collapseChildren(top);
        tree.collapsePath(new TreePath(top.getPath()));
        view.clear(false);
    }    
    
    public void collapseChildren(DefaultMutableTreeNode parent) {
        int size = parent.getChildCount();
        for (int i = 0; i < size; i++) {
            try {
                DefaultMutableTreeNode child = (DefaultMutableTreeNode) parent.getChildAt(i);
                tree.collapsePath(new TreePath(child.getPath()));
            } catch (Exception exp) {
                Messages.add(exp);
            }
        }
    }
    
    public void expandAll() {
        expandAll(top);
    }
    
    private void expandAll(DefaultMutableTreeNode node) {
        try {
            DefaultTreeModel model = (DefaultTreeModel) tree.getModel();
            int count = node.getChildCount();
            tree.expandPath(new TreePath(model.getPathToRoot(node)));
            for (int i = 0; i < count; i++) {
                DefaultMutableTreeNode current = (DefaultMutableTreeNode) node.getChildAt(i);
                tree.expandPath(new TreePath(model.getPathToRoot(current)));
                expandAll(current);
            }
        } catch (Exception exp) {}
    }
    
    public boolean isActive() {
        boolean active = isEnabled() && isVisible() && top != null && fields != null && fields.length > 0;
        
        if (active) {
            for (int i = 0; i < fields.length; i++)
                active &= fields[i] != -1;
        }
            
        return active;
    }
    
    public Collection getObjects() {
        Collection c = new ArrayList();
        c.addAll(objects);
        return c;
    }
    
    private synchronized void addLeafs(DefaultMutableTreeNode parentNode, Collection objects, int level) {
        int fldIdx = fields[level];
        
        if (DcModules.get(module).getField(fldIdx).getValueType() == Repository.datatypes._DCOBJECTCOLLECTION) {
            DcField fld = DcModules.get(module).getField(fldIdx);
            DcObject[] c = DataManager.get(((DcMappingModule) DcModules.get(fld.getModRef())).getSimpleModuleIdx(), null);
            
            for (int i = 0; c != null &&  i < c.length; i++) {
                NodeElement ne = new NodeElement();
                ne.setKey(c[i].toString());
                parentNode.add(new DefaultMutableTreeNode(ne));
            }
            
            for (Iterator iter = objects.iterator(); iter.hasNext(); ) {
                DcObject dco = (DcObject) iter.next();
                addElement(dco, false);
            }

            removeEmptyLeafs(parentNode);
            
        } else {
            Collections.sort((List) objects, new DcObjectComparator(fldIdx));
            NodeElement currentNode = new NodeElement();
            for (Iterator iter = objects.iterator(); iter.hasNext(); ) {
                DcObject dco = (DcObject) iter.next();
                
                if (view != null) {
                    view.setStatus(DcResources.getText("msgAddingXToView", dco.toString()));
                    view.updateProgressBar();
                }
                
                String key = dco.getDisplayString(fldIdx);
                if (!key.equals(currentNode.getKey())) {
                    currentNode = new NodeElement();
                    currentNode.setKey(key);
                    parentNode.add(new DefaultMutableTreeNode(currentNode));
                }
            
                currentNode.addValue(dco);
            }
        }
    }
    
    private synchronized void buildTree(DcObject[] o) {
        clear();

        objects = new ArrayList();
        for (int i = 0; i < o.length; i++)
            objects.add(o[i]);

        buildTree(objects);
    }
    
    public synchronized void buildTree(Collection c) {
        
        initializeTree();

        if (!isActive()) {
            NodeElement node = new NodeElement();
            node.setKey("all");
            node.setValues(objects);
            top.add(new DefaultMutableTreeNode(node));
        } else {
            addLeafs(top, objects, 0);

            for (int level = 1; level < fields.length; level++) {
                int size = top.getChildCount();
                for (int pos = 0; pos < size; pos++) {
                    DefaultMutableTreeNode parentNode = (DefaultMutableTreeNode) top.getChildAt(pos);
                    NodeElement ne = (NodeElement) parentNode.getUserObject();
                    addLeafs(parentNode, ne.getValues(), level);
                }
            }
        }

        expandAll();
        tree.setSelectionInterval(1, 1);
    }
    
    public void reset() {
        DcSettings.setValue(DcModules.get(module).getGroupedBySettingsKey(), null);
        fields = null;
        Collection objects = getObjects();
        clear();
        initializeTree();
        
        DcObject[] o = new DcObject[objects.size()];
        add((DcObject[]) objects.toArray(o));
    }
    
    private void clear() {
        if (tree != null) {
            objects.clear();
            
            tree.removeTreeSelectionListener(tsl);
            tree = null;
            
            top.removeAllChildren();
            remove(scroller);
            top = null;
            
            revalidate();
            repaint();
        }
    }
    
    private void initializeTree() {
        if (tree != null) 
            return;
            
        DcModule mod = DcModules.get(module);
        String orderingOn = mod.getObjectNamePlural();
        
        if (isActive()) {
            orderingOn += " by ";
            
            for (int i = 0; i < fields.length; i++)
                orderingOn += (i > 0 ? " & " : "") + mod.getField(fields[i]).getLabel();
        }
        
        top = new DefaultMutableTreeNode(orderingOn);
        tree = new DcTree(new DefaultTreeModel(top));
        tree.setFont(ComponentFactory.getStandardFont());
        tree.addTreeSelectionListener(tsl);
        
        scroller = new JScrollPane(tree);
        
        add(scroller,  Layout.getGBC( 0, 1, 1, 1, 50.0, 50.0
           ,GridBagConstraints.NORTHWEST, GridBagConstraints.BOTH,
            new Insets(0, 5, 5, 5), 0, 0));
        
        this.revalidate();
        this.repaint();
    }
    
    private void initializeMenu() {
        
        add(menu,  Layout.getGBC( 0, 0, 1, 1, 1.0, 1.0
                  ,GridBagConstraints.NORTHWEST, GridBagConstraints.HORIZONTAL,
                   new Insets(5, 5, 0, 5), 0, 0));
    }
    
    private class DcTreeSelectionListener implements TreeSelectionListener {
        public void valueChanged(TreeSelectionEvent e) {
            try {
                if (e.getNewLeadSelectionPath().equals(e.getOldLeadSelectionPath()))
                    return;
                
                DefaultMutableTreeNode node = (DefaultMutableTreeNode) tree.getLastSelectedPathComponent();
    
                if (node == null) {
                    currentUserObject = null;
                    return;
                }
    
                Object o =  node.getUserObject();
                
                if (!o.equals(currentUserObject)) {
                    currentUserObject =  o;
                    if (currentUserObject instanceof NodeElement) {
                        view.clear(clearSafely);
                        NodeElement currentNode = (NodeElement) o;
                        view.add(currentNode.getValues(), false);
                    } else {
                        view.clear(clearSafely);
                    }
                }
            } catch (Exception ignore) {}
        }
    }  
    
    private class NodeElement {
        
        private String key;
        private Collection values = new ArrayList();
        
        public NodeElement() {
        }

        public void setKey(String key) {
            this.key = key;
        }

        public void setValues(Collection values) {
            this.values = values;
        }
        
        public void addValue(DcObject dco) {
            if (!values.contains(dco))
                values.add(dco);
        }
        
        public void updateValue(DcObject dco) {
            int pos = ((List) values).indexOf(dco);
            if (pos != -1) {
                DcObject o = (DcObject) ((List) values).get(pos);
                o.reload();
            }
        }
        
        public int hashCode() {
            return key.hashCode();
        }
        
        public void removeValue(DcObject dco) {
            values.remove(dco);
        }
        
        public Collection getValues() {
            if (values != null) {
                DataFilter df = DataFilters.getDefaultDataFilter(module);
                df.sort(values);
            }
            return values;
        }
        
        public String getKey() {
            return key;
        }
        
        public String toString() {
            return key + " (" + values.size() + ")";
        }
    }
    
    private class DcObjectComparator implements Comparator {

        private final int field;
        
        public DcObjectComparator(int field) {
            this.field = field;
        }
        
        public int compare(Object o1, Object o2) {
            DcObject dco1 = (DcObject) o1;
            DcObject dco2 = (DcObject) o2;
            
            String value1 = dco1.getDisplayString(field);
            String value2 = dco2.getDisplayString(field);
        
            try {
                int num1 = Integer.parseInt(value1);
                int num2 = Integer.parseInt(value2);
                return num1 - num2;
            } catch (Exception exp) {
                return value1.compareTo(value2); 
            }
        }
    }
}
