/******************************************************************************
 *                                     __                                     *
 *                              <-----/@@\----->                              *
 *                             <-< <  \\//  > >->                             *
 *                               <-<-\ __ /->->                               *
 *                               Data /  \ Crow                               *
 *                                   ^    ^                                   *
 *                       (c) 2003 The Data Crow team                          *
 *                              info@datacrow.net                             *
 *                                                                            *
 *                                                                            *
 *       This library is free software; you can redistribute it and/or        *
 *        modify it under the terms of the GNU Lesser General Public          *
 *       License as published by the Free Software Foundation; either         *
 *     version 2.1 of the License, or (at your option) any later version.     *
 *                                                                            *
 *      This library is distributed in the hope that it will be useful,       *
 *      but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU       *
 *           Lesser General Public License for more details.                  *
 *                                                                            *
 *     You should have received a copy of the GNU Lesser General Public       *
 *    License along with this library; if not, write to the Free Software     *
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA   *
 *                                                                            *
 ******************************************************************************/

package net.sf.dc.console.dialogs;

import java.awt.Dimension;
import java.awt.GridBagConstraints;
import java.awt.Insets;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.awt.event.WindowAdapter;
import java.awt.event.WindowEvent;
import java.io.BufferedInputStream;
import java.io.File;
import java.io.FileInputStream;
import java.io.FileOutputStream;
import java.io.InputStream;
import java.text.SimpleDateFormat;
import java.util.ArrayList;
import java.util.Calendar;
import java.util.Collection;
import java.util.Enumeration;
import java.util.Iterator;
import java.util.zip.ZipEntry;
import java.util.zip.ZipFile;
import java.util.zip.ZipOutputStream;

import javax.swing.JButton;
import javax.swing.JPanel;
import javax.swing.JProgressBar;
import javax.swing.JScrollPane;
import javax.swing.JTextArea;
import javax.swing.JToolTip;
import javax.swing.filechooser.FileFilter;

import net.sf.dc.console.ComponentFactory;
import net.sf.dc.console.Layout;
import net.sf.dc.console.components.DcDialog;
import net.sf.dc.console.components.DcFileField;
import net.sf.dc.console.components.DcMultiLineToolTip;
import net.sf.dc.core.DataCrow;
import net.sf.dc.core.Repository;
import net.sf.dc.core.modules.DcModule;
import net.sf.dc.core.modules.DcModules;
import net.sf.dc.db.DatabaseManager;
import net.sf.dc.messages.Messages;
import net.sf.dc.reporting.templates.ReportTemplates;
import net.sf.dc.resources.DcResources;
import net.sf.dc.settings.DcSettings;
import net.sf.ext.util.DirectoryReader;
import net.sf.ext.util.Utilities;

public class BackupDialog extends DcDialog {

    private final DcFileField fileFieldTarget;
    private final DcFileField fileFieldSource;

    private final JButton buttonBackup = ComponentFactory.getButton(DcResources.getText("lblBackup"));
    private final JButton buttonRestore = ComponentFactory.getButton(DcResources.getText("lblRestore"));

    private final JTextArea textLog = ComponentFactory.getTextArea();
    private final JProgressBar progressBar = new JProgressBar();

    public BackupDialog() {
        super(DataCrow.mainFrame);
        setTitle(DcResources.getText("lblBackupAndRestore"));
        setHelpIndex("dc.Tools.Backup");

        BackupFileFilter filter = new BackupFileFilter();
        fileFieldSource = ComponentFactory.getFileField(false, false, filter);
        fileFieldTarget = ComponentFactory.getFileField(false, true);

        addWindowListener(new WindowAdapter() {
            public void windowClosing(WindowEvent e) {
                close();
            }
        });

        buildDialog();
    }

    private Collection getFiles() {
        File dataDir = new File(DataCrow.baseDir + "data" + File.separator);
        String[] list = dataDir.list();
        Collection files = new ArrayList();
        for (int i = 0; i < list.length; i++) {
            String file = list[i];
            File fl = new File(DataCrow.baseDir + "data" + File.separator + file);
            if (!fl.isDirectory() && !file.endsWith(".log") && !file.equals("images"))
                files.add(DataCrow.baseDir + "data" + File.separator + file);
        }
        
        files.addAll(
            DirectoryReader.read(DataCrow.imageDir, false, true, false, new String[] {"jpg", "jpeg"}, false));        
        
        for (Iterator iter = new ReportTemplates().getFolders().iterator(); iter.hasNext(); ) {
            files.addAll(DirectoryReader.read((String) iter.next(), false, true, false, new String[] {"xsl", "xslt"}, false));
        }
        
        initProgressBar(files.size());
        
        return files;
    }

    private ZipOutputStream getZipOutputStream(String target) {
        Calendar cal = Calendar.getInstance();
        SimpleDateFormat format = new SimpleDateFormat("yyyyMMdd");
        String date = format.format(cal.getTime());

        String filename = "datacrow_backup_" + date + ".bck";
        String zipFile = target.endsWith(File.separator) ? target + filename :
                                                           target + File.separator + filename;
        ZipOutputStream zout = null;
        try {
            FileOutputStream fos = new FileOutputStream(zipFile);
            zout = new ZipOutputStream(fos);
        } catch (Exception exp) {
            new MessageBox(DcResources.getText("msgBackupFileCreationError", exp.getMessage()), MessageBox._ERROR);
            Messages.add(exp);
        }

        return zout;
    }

    private void close() {
        DcSettings.setValue(Repository.settings.stBackupDialogSize, getSize());
        dispose();
    }

    private void addMessage(String message) {
        textLog.insert(message + '\n', 0);
        textLog.setCaretPosition(0);
    }

    public void initProgressBar(int maxValue) {
        progressBar.setValue(0);
        progressBar.setMaximum(maxValue);
    }

    public void updateProgressBar(int value) {
        progressBar.setValue(value);
    }

    private void restartApplication() {
        initProgressBar(3);
        addMessage(DcResources.getText("msgRestartingDb"));
        DatabaseManager.initialize();
        updateProgressBar(1);

        addMessage(DcResources.getText("msgReapplyingSettings"));
        new DcSettings();
        new DcResources();
        DataCrow.mainFrame.applySettings();
        DataCrow.mainFrame.updateLAF(DcSettings.getValueAsLookAndFeel(Repository.settings.stLookAndFeel));
        updateProgressBar(2);
        
        addMessage("Loading items");
        DcModules.loadData();
        updateProgressBar(3);
        
        DataCrow.mainFrame.changeModule(DcSettings.getValueAsInt(Repository.settings.stModule), false);
        for (Iterator iter = DcModules.getAllModules(); iter.hasNext(); ) {
            DcModule module = (DcModule) iter.next();
            if (module.getSearchView() != null)
                module.getSearchView().clear();    
        }
        
        DataCrow.mainFrame.bindData();
    }

    private void buildDialog() {
        getContentPane().setLayout(Layout.getGBL());

        //**********************************************************
        //Create Backup Panel
        //**********************************************************
        JPanel panelBackup = new JPanel() {
            public JToolTip createToolTip() {
                return new DcMultiLineToolTip();
            }
        };

        panelBackup.setLayout(Layout.getGBL());

        buttonBackup.addActionListener(new BackupAction());

        panelBackup.add(fileFieldTarget, Layout.getGBC( 0, 0, 1, 1, 1.0, 1.0
                ,GridBagConstraints.NORTHWEST, GridBagConstraints.HORIZONTAL,
                 new Insets( 0, 5, 5, 5), 0, 0));
        panelBackup.add(buttonBackup,  Layout.getGBC( 1, 0, 1, 1, 0.0, 0.0
                ,GridBagConstraints.NORTHWEST, GridBagConstraints.NONE,
                 new Insets( 0, 0, 5, 5), 0, 0));
        panelBackup.setToolTipText(DcResources.getText("tpBackup"));

        panelBackup.setBorder(
                ComponentFactory.getTitleBorder(DcResources.getText("lblBackupHeader")));

        //**********************************************************
        //Restore Restore Panel
        //**********************************************************
        JPanel panelRestore = new JPanel()  {
            public JToolTip createToolTip() {
                return new DcMultiLineToolTip();
            }
        };

        panelRestore.setLayout(Layout.getGBL());
        buttonRestore.addActionListener(new RestoreAction());

        panelRestore.add(fileFieldSource, Layout.getGBC( 0, 0, 1, 1, 1.0, 1.0
                ,GridBagConstraints.NORTHWEST, GridBagConstraints.HORIZONTAL,
                 new Insets( 0, 5, 5, 5), 0, 0));
        panelRestore.add(buttonRestore,  Layout.getGBC( 1, 0, 1, 1, 0.0, 0.0
                ,GridBagConstraints.NORTHWEST, GridBagConstraints.NONE,
                 new Insets( 0, 0, 5, 5), 0, 0));

        panelRestore.setToolTipText(DcResources.getText("tpRestore"));

        panelRestore.setBorder(
                ComponentFactory.getTitleBorder(DcResources.getText("lblRestoreHeader")));

        //**********************************************************
        //Log Panel
        //**********************************************************
        JPanel panelLog = new JPanel();
        panelLog.setLayout(Layout.getGBL());

        JScrollPane scroller = new JScrollPane(textLog);
        scroller.setHorizontalScrollBarPolicy(JScrollPane.HORIZONTAL_SCROLLBAR_NEVER);
        panelLog.add(scroller, Layout.getGBC( 0, 1, 1, 1, 1.0, 1.0
                    ,GridBagConstraints.NORTHWEST, GridBagConstraints.BOTH,
                     new Insets(5, 5, 5, 5), 0, 0));

        panelLog.setBorder(ComponentFactory.getTitleBorder(DcResources.getText("lblLog")));

        //**********************************************************
        //Progress panel
        //**********************************************************
        JPanel panelProgress = new JPanel();
        panelProgress.setLayout(Layout.getGBL());
        panelProgress.add(progressBar, Layout.getGBC( 0, 1, 1, 1, 1.0, 1.0
                         ,GridBagConstraints.NORTHWEST, GridBagConstraints.HORIZONTAL,
                          new Insets(5, 5, 5, 5), 0, 0));


        //**********************************************************
        //Main panel
        //**********************************************************
        this.getContentPane().setLayout(Layout.getGBL());

        JButton buttonCancel = ComponentFactory.getButton(DcResources.getText("lblCancel"));

        buttonCancel.addActionListener(new CloseAction());
        buttonCancel.setMnemonic('C');
        buttonBackup.setMnemonic('B');
        buttonRestore.setMnemonic('R');

        this.getContentPane().add(panelBackup,  Layout.getGBC( 0, 0, 1, 1, 1.0, 1.0
                                 ,GridBagConstraints.NORTHWEST, GridBagConstraints.HORIZONTAL,
                                  new Insets( 5, 5, 5, 5), 0, 0));
        this.getContentPane().add(panelRestore, Layout.getGBC( 0, 1, 1, 1, 1.0, 1.0
                                 ,GridBagConstraints.NORTHWEST, GridBagConstraints.HORIZONTAL,
                                  new Insets( 5, 5, 5, 5), 0, 0));
        this.getContentPane().add(panelLog,     Layout.getGBC( 0, 2, 1, 1, 10.0, 10.0
                                 ,GridBagConstraints.NORTHWEST, GridBagConstraints.BOTH,
                                  new Insets( 5, 5, 5, 5), 0, 0));
        this.getContentPane().add(panelProgress,Layout.getGBC( 0, 3, 1, 1, 1.0, 1.0
                                 ,GridBagConstraints.NORTHWEST, GridBagConstraints.HORIZONTAL,
                                  new Insets( 5, 5, 5, 5), 0, 0));
        this.getContentPane().add(buttonCancel,  Layout.getGBC( 0, 4, 1, 1, 0.0, 0.0
                                 ,GridBagConstraints.NORTHEAST, GridBagConstraints.NONE,
                                  new Insets( 5, 5, 5, 10), 0, 0));
        pack();

        Dimension size = DcSettings.getValueAsDimension(Repository.settings.stBackupDialogSize);
        setSize(size);

        setCenteredLocation();
    }

    private class BackupAction implements ActionListener {
        public void actionPerformed(ActionEvent e) {
            BackupThread thread = new BackupThread();
            thread.start();
        }
    }

    private class RestoreAction implements ActionListener {
        public void actionPerformed(ActionEvent e) {
            RestoreThread thread = new RestoreThread();
            thread.start();
        }
    }

    private class CloseAction implements ActionListener {
        public void actionPerformed(ActionEvent e) {
            close();
        }
    }

    /**
     * Image file filter
     */
    private class BackupFileFilter extends FileFilter {

        public boolean accept(File file) {
            if (file.isDirectory()) {
                return true;
            } else if (Utilities.getExtension(file).equals("bck")) {
                return true;
            } else {
                return false;
            }
        }

        public String getDescription() {
            return DcResources.getText("lblBckFileFilter");
        }
    }

    private class BackupThread extends Thread {
        public void run() {
            String target = fileFieldTarget.getFileAsString();

            File tmp = new File(target);
            if (!tmp.exists())
                tmp.mkdirs();
            
            textLog.setText("");
            if (target != null && target.length() > 0) {
                addMessage(DcResources.getText("msgStartBackup"));

                buttonBackup.setEnabled(false);
                buttonRestore.setEnabled(false);

                DatabaseManager.closeDatabases(false);
                addMessage(DcResources.getText("msgClosingDb"));

                Collection files = getFiles();

                try {
                    byte b[] = new byte[512];
                    ZipOutputStream zout = getZipOutputStream(target);

                    if (zout != null) {
                        int counter = 1;
                        for (Iterator iter = files.iterator(); iter.hasNext(); counter++) {
                            String file = (String) iter.next();

                            if (!file.endsWith(".log")) {
                                InputStream in = new FileInputStream(file);
                                ZipEntry e = new ZipEntry(file.replace(File.separatorChar, '/'));
                                zout.putNextEntry(e);
                                int len = 0;
                                while ((len = in.read(b)) != -1) {
                                    zout.write(b, 0, len);
                                }

                                addMessage(DcResources.getText("msgCreatingBackupOfFile", file));

                                zout.closeEntry();
                            }

                            updateProgressBar(counter);
                        }
                        zout.close();

                        addMessage(DcResources.getText("msgWritingBackupFile"));
                    }
                } catch (Exception exp) {
                    addMessage(DcResources.getText("msgBackupError", exp.getMessage()));
                    Messages.add(exp);
                }

                addMessage(DcResources.getText("msgRestartingDb"));
                DatabaseManager.initialize();
                addMessage(DcResources.getText("msgBackupFinished"));
            } else {
                new MessageBox(DcResources.getText("msgSelectOutputDir"), MessageBox._WARNING);
            }
            buttonRestore.setEnabled(true);
            buttonBackup.setEnabled(true);
        }

    }

    private class RestoreThread extends Thread {

        public void run() {

            String source = fileFieldSource.getFileAsString();

            textLog.setText("");
            if (source != null && source.length() > 0) {
                addMessage(DcResources.getText("msgStartRestore"));

                buttonRestore.setEnabled(false);
                buttonBackup.setEnabled(false);

                addMessage(DcResources.getText("msgClosingDb"));
                DatabaseManager.closeDatabases(false);

                try {
                    ZipFile zf = new ZipFile(source);

                    Enumeration list = zf.entries();
                    initProgressBar(zf.size());
                    int counter = 1;
                    while (list.hasMoreElements()) {
                        ZipEntry ze = (ZipEntry)list.nextElement();

                        String filename = ze.getName();
                        
                        if (filename.toLowerCase().endsWith(".jpg")) {
                            filename = DataCrow.imageDir +
                                       filename.substring(filename.lastIndexOf("/") + 1, filename.length());
                        } else if (filename.toLowerCase().endsWith(".xsl") || filename.toLowerCase().endsWith(".xslt") ||
                                   filename.toLowerCase().indexOf("/reports") > -1) {
                            filename = filename.substring(filename.lastIndexOf("/reports") + 1, filename.length());
                            filename = DataCrow.baseDir + filename;
                        } else {
                            filename = DataCrow.baseDir + "data/" + 
                                       filename.substring(filename.lastIndexOf("/") + 1, filename.length());
                        }
                        
                        InputStream istr = zf.getInputStream(ze);
                        BufferedInputStream bis = new BufferedInputStream(istr);
                        FileOutputStream fos = new FileOutputStream(filename);

                        int sz = (int)ze.getSize();
                        final int N = 1024;
                        byte buf[] = new byte[N];
                        int ln = 0;
                        while (sz > 0 &&  // workaround for bug
                            (ln = bis.read(buf, 0, Math.min(N, sz))) != -1) {
                                fos.write(buf, 0, ln);
                                sz -= ln;
                        }
                        bis.close();
                        fos.flush();
                        fos.close();

                        addMessage(DcResources.getText("msgRestoringFile", filename));
                        updateProgressBar(counter++);
                    }

                } catch (Exception exp) {
                    addMessage(DcResources.getText("msgRestoreError", exp.getMessage()));
                    new MessageBox(DcResources.getText("msgRestoreError", exp.getMessage()), MessageBox._WARNING);
                    Messages.add(exp);
                }

                restartApplication();

                addMessage(DcResources.getText("msgRestoreFinished"));

             } else {
                 new MessageBox(DcResources.getText("msgSelectBackupFile"), MessageBox._WARNING);
             }
             buttonBackup.setEnabled(true);
             buttonRestore.setEnabled(true);
        }
    }
}
