package net.sf.dc.console.forms;

import java.awt.GridBagConstraints;
import java.awt.Insets;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.awt.event.WindowAdapter;
import java.awt.event.WindowEvent;
import java.util.Collection;

import javax.swing.JButton;
import javax.swing.JCheckBox;
import javax.swing.JPanel;
import javax.swing.JScrollPane;
import javax.swing.JTextArea;

import net.sf.dc.console.ComponentFactory;
import net.sf.dc.console.Layout;
import net.sf.dc.console.components.DcDialog;
import net.sf.dc.console.components.DcLongTextField;
import net.sf.dc.console.components.DcProgressBar;
import net.sf.dc.console.components.panels.OnlineServicePanel;
import net.sf.dc.console.components.panels.OnlineServiceSettingsPanel;
import net.sf.dc.core.DataCrow;
import net.sf.dc.core.Repository;
import net.sf.dc.core.modules.DcModule;
import net.sf.dc.core.services.DcRegion;
import net.sf.dc.core.services.DcServer;
import net.sf.dc.messages.Messages;
import net.sf.dc.processes.synchronizer.Synchronizer;
import net.sf.dc.resources.DcResources;
import net.sf.dc.settings.DcSettings;

public class ItemSynchronizerDialog extends DcDialog {
    
    private DcProgressBar progressBar = new DcProgressBar();
    private JTextArea textLog = ComponentFactory.getTextArea();
    private JButton buttonStart = ComponentFactory.getButton(DcResources.getText("lblRun"));
    private JButton buttonStop = ComponentFactory.getButton(DcResources.getText("lblStop"));
    private JButton buttonCancel = ComponentFactory.getButton(DcResources.getText("lblCancel"));
    private DcLongTextField textHelp = ComponentFactory.getLongTextField(8000);
    
    private JCheckBox checkReparseFiles = ComponentFactory.getCheckBox(DcResources.getText("lblReparseMusicFiles"));
    
    private OnlineServiceSettingsPanel panelSettings;
    private OnlineServicePanel panelServer;
    private Synchronizer synchronizer;
    
    private final int module;
    private boolean cancelled = true;
    private final boolean canParseFiles;
    
    public ItemSynchronizerDialog(DcModule module) {
        super(DataCrow.mainFrame);

        this.module = module.getIndex();
        this.synchronizer = module.getSynchronizer();
        this.canParseFiles = synchronizer.canParseFiles();

        try {
            setTitle(synchronizer.getTitle());
            setHelpIndex(synchronizer.getHelpIndex());
    
            buildDialog(module.getOnlineServices());
            
            
            addWindowListener(new WindowAdapter() {
                public void windowClosing(WindowEvent e) {
                    close();
                }
            });     
            
            pack();
            setSize(DcSettings.getValueAsDimension(Repository.settings.stItemSynchronizerDialogSize));
            setCenteredLocation();
            
            enableActions(true);
        } catch (Exception exp) {
            Messages.add(exp);
        }
    }
    
    protected void synchronize() {
        synchronizer.synchronize(this);
    }
    
    public boolean isReparseFiles() {
        return checkReparseFiles.isSelected();
    }
    
    public boolean useOnlineService() {
        return panelServer.useOnlineService();
    }

    public void initialize() {
        cancelled = false;
        initProgressBar(0);
        enableActions(false);
    }
    
    protected void cancel() {
        cancelled = true;
        enableActions(true);
    }
    
    public boolean isCancelled() {
        return cancelled;
    }
    
    private void close() {
        cancel();
        
        cancelled = true;
        
        progressBar = null;
        textLog = null;
        buttonStart = null;
        buttonCancel = null;
        textHelp = null;
        buttonStop = null;
        
        panelSettings.save();
        panelSettings.clear();
        panelSettings = null;
        
        DcSettings.setValue(Repository.settings.stItemSynchronizerDialogSize, getSize());

        dispose();
        setVisible(false);
    }    
    
    public DcServer getServer() {
        return panelServer.getServer();
    }

    public DcRegion getRegion() {
        return panelServer.getRegion();
    }
    
    public void enableActions(boolean b) {
        if (buttonStart != null)
            buttonStart.setEnabled(b);
        
        if (buttonStop != null)
            buttonStop.setEnabled(!b);
        
        if (!b) {
            progressBar.setValue(0);
            textLog.setText("");
        }
    }    
    
    public void addMessage(String message) {
        if (textLog != null) 
            textLog.insert(message + '\n', 0);
    }

    public void initProgressBar(int maxValue) {
        progressBar.setValue(0);
        progressBar.setMaximum(maxValue);
    }

    public void updateProgressBar() {
        if (progressBar != null)
            progressBar.setValue(progressBar.getValue() + 1);
    }
    
    private void buildDialog(Collection servers) {
        getContentPane().setLayout(Layout.getGBL());

        //**********************************************************
        //Help panel
        //**********************************************************
        textHelp.setText(synchronizer.getHelpText());
        textHelp.setBorder(null);

        JScrollPane helpScroller = new JScrollPane(textHelp);
        helpScroller.setBorder(null);
        helpScroller.setHorizontalScrollBarPolicy(JScrollPane.HORIZONTAL_SCROLLBAR_NEVER);
        helpScroller.setVerticalScrollBarPolicy(JScrollPane.VERTICAL_SCROLLBAR_NEVER);
        textHelp.setEditable(false);
        
        //**********************************************************
        //Online Server panel
        //**********************************************************
        panelServer = new OnlineServicePanel(servers, canParseFiles ? true : false);

        //**********************************************************
        //Online Server Settings panel
        //**********************************************************
        panelSettings = new OnlineServiceSettingsPanel(module, true);
        
        //**********************************************************
        //Reparse panel
        //**********************************************************
        JPanel panelReparse = new JPanel();
        
        if (canParseFiles) {
            panelReparse.setLayout(Layout.getGBL());
            panelReparse.setBorder(ComponentFactory.getTitleBorder(DcResources.getText("lblMusicFileProcessingConfig")));
            
            panelReparse.add(checkReparseFiles, Layout.getGBC(0, 0, 1, 1, 1.0, 1.0
                            ,GridBagConstraints.NORTHWEST, GridBagConstraints.HORIZONTAL,
                             new Insets(5, 5, 5, 5), 0, 0));
            checkReparseFiles.setSelected(true);
        }
        
        //**********************************************************
        //Log panel
        //**********************************************************
        JPanel panelLog = new JPanel();
        panelLog.setLayout(Layout.getGBL());
        
        JScrollPane logScroller = new JScrollPane(textLog);
        logScroller.setHorizontalScrollBarPolicy(JScrollPane.HORIZONTAL_SCROLLBAR_NEVER);

        panelLog.setBorder(ComponentFactory.getTitleBorder(DcResources.getText("lblLog")));
        panelLog.add(logScroller, Layout.getGBC(0, 0, 1, 1, 1.0, 1.0
                    ,GridBagConstraints.NORTHWEST, GridBagConstraints.BOTH,
                    new Insets(5, 5, 5, 5), 0, 0));


        //**********************************************************
        //Action panel
        //**********************************************************
        JPanel panelActions = new JPanel();
        
        buttonStart.addActionListener(new StartViewUpdate());
        buttonCancel.addActionListener(new CancelAction());
        buttonStop.addActionListener(new StopAction());
        
        panelActions.add(buttonStart);
        panelActions.add(buttonStop);
        panelActions.add(buttonCancel);
        
        buttonStart.setMnemonic('S');
        buttonStop.setMnemonic('T');
        buttonCancel.setMnemonic('C');  
        
        //**********************************************************
        //Main
        //**********************************************************
        getContentPane().add(helpScroller,      Layout.getGBC(0, 0, 1, 1, 1.0, 1.0
                ,GridBagConstraints.NORTHWEST, GridBagConstraints.BOTH,
                 new Insets(5, 5, 5, 5), 0, 0));
        getContentPane().add(panelServer,       Layout.getGBC(0, 1, 1, 1, 1.0, 1.0
                ,GridBagConstraints.NORTHWEST, GridBagConstraints.BOTH,
                 new Insets(5, 5, 5, 5), 0, 0));
        getContentPane().add(panelReparse,      Layout.getGBC(0, 2, 1, 1, 1.0, 1.0
                ,GridBagConstraints.NORTHWEST, GridBagConstraints.HORIZONTAL,
                 new Insets(5, 5, 5, 5), 0, 0));
        getContentPane().add(panelSettings,     Layout.getGBC(0, 3, 1, 1, 1.0, 1.0
                ,GridBagConstraints.NORTHWEST, GridBagConstraints.BOTH,
                 new Insets(5, 5, 5, 5), 0, 0));
        getContentPane().add(panelActions,      Layout.getGBC( 0, 4, 1, 1, 1.0, 1.0
                ,GridBagConstraints.NORTHEAST, GridBagConstraints.NONE,
                 new Insets(0, 0, 0, 0), 0, 0));
        getContentPane().add(panelLog,          Layout.getGBC( 0, 5, 1, 1, 5.0, 5.0
                ,GridBagConstraints.SOUTHWEST, GridBagConstraints.BOTH,
                 new Insets(0, 0, 0, 0), 0, 0));
        getContentPane().add(progressBar,       Layout.getGBC( 0, 6, 1, 1, 1.0, 1.0
                ,GridBagConstraints.SOUTHWEST, GridBagConstraints.HORIZONTAL,
                 new Insets(0, 0, 0, 0), 0, 0));
    }

    private class CancelAction implements ActionListener {
        public void actionPerformed(ActionEvent e) {
            close();
        }
    }

    private class StopAction implements ActionListener {
        public void actionPerformed(ActionEvent e) {
            cancel();
        }
    }
    
    private class StartViewUpdate implements ActionListener {
        public void actionPerformed(ActionEvent e) {
            synchronize();
        }
    }

}
