/******************************************************************************
 *                                     __                                     *
 *                              <-----/@@\----->                              *
 *                             <-< <  \\//  > >->                             *
 *                               <-<-\ __ /->->                               *
 *                               Data /  \ Crow                               *
 *                                   ^    ^                                   *
 *                       (c) 2003 The Data Crow team                          *
 *                              info@datacrow.net                             *
 *                                                                            *
 *                                                                            *
 *       This library is free software; you can redistribute it and/or        *
 *        modify it under the terms of the GNU Lesser General Public          *
 *       License as published by the Free Software Foundation; either         *
 *     version 2.1 of the License, or (at your option) any later version.     *
 *                                                                            *
 *      This library is distributed in the hope that it will be useful,       *
 *      but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU       *
 *           Lesser General Public License for more details.                  *
 *                                                                            *
 *     You should have received a copy of the GNU Lesser General Public       *
 *    License along with this library; if not, write to the Free Software     *
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA   *
 *                                                                            *
 ******************************************************************************/

package net.sf.dc.console.views;

import java.util.ArrayList;
import java.util.Collection;
import java.util.Collections;
import java.util.Iterator;

import javax.swing.ImageIcon;
import javax.swing.JComponent;

import net.sf.dc.console.components.DcPanel;
import net.sf.dc.console.components.DcViewDivider;
import net.sf.dc.console.components.panels.TreePanel;
import net.sf.dc.console.dialogs.MessageBox;
import net.sf.dc.console.dialogs.UpdateAllDialog;
import net.sf.dc.console.forms.ItemForm;
import net.sf.dc.core.Repository;
import net.sf.dc.core.data.ResultSorter;
import net.sf.dc.core.modules.DcModule;
import net.sf.dc.core.objects.DcField;
import net.sf.dc.core.objects.DcObject;
import net.sf.dc.processes.DataTask;
import net.sf.dc.resources.DcResources;
import net.sf.dc.settings.DcSettings;
import net.sf.ext.util.CompositeComparator;

public abstract class View extends DcPanel {

    protected TreePanel treePanel;
    
    protected DataTask task;
    private final int index; 
    private final DcModule module;
    
    protected DcViewDivider vdQuickPane;
    protected DcViewDivider vdGroupingPane;
    
    private boolean overwriteVisibility = false;
    
    public View(DcModule module, String title, ImageIcon icon, int index) {
        super(title, icon);  
        this.index = index;
        this.module = module;
    }    
    
    protected boolean isTaskRunning() {
        boolean isTaskRunning = task != null && task.isRunning();
        if (isTaskRunning)
            new MessageBox(DcResources.getText("msgJobRunning"), MessageBox._WARNING); 
        
        return isTaskRunning; 
    }
    
    public void setOverwriteVisibility(boolean b) {
        overwriteVisibility = b;
    }
    
    public boolean isVisibilityOverwritten() {
        return overwriteVisibility;
    }
    
    public void applyViewDividerLocation() {
        if (vdQuickPane != null)
            vdQuickPane.setDividerLocation(DcSettings.getValueAsInt(Repository.settings.stQuickViewDividerLocation));
        
        if (vdGroupingPane != null)
            vdGroupingPane.setDividerLocation(DcSettings.getValueAsInt(Repository.settings.stTreeDividerLocation));
    }
    
    public DataTask getCurrentTask() {
        return task;
    }
    
    public void markAsUnchanged() {
    }
    
    public int getIndex() {
        return index;
    }
    
    public void sort() {
        String[] s = DcSettings.getValueAsStringArray(module.getSearchOrderSettingsKey());
        DcField[] fields = new DcField[s.length];
        
        Collection c = module.getFields();
        int counter = 0;
        for (Iterator iter = c.iterator(); iter.hasNext(); ) {
            DcField field = (DcField) iter.next();
            for (int i = 0; i < s.length; i++) {
                if (field.getDatabaseFieldName().equals(s[i]))
                    fields[counter++] = field;
            }
        }
        
        int rowCount = getRows();
        ArrayList objects = new ArrayList();
        for (int i = 0; i < rowCount; i++) {
            objects.add(getObject(i));
        }
        
        if (fields.length == 1)
            Collections.sort(objects, new ResultSorter(fields[0]));
        else if (fields.length == 2)
            Collections.sort(objects, 
                new CompositeComparator(new ResultSorter(fields[0]), new ResultSorter(fields[1])));   
        
        add(objects, false);
    }
    
    public void open(DcObject dco, boolean update, boolean applyTemplate) {
        dco.setRow(-1);
        ItemForm itemForm = new ItemForm(    null,
                                             false,
                                             update,
                                             dco,
                                             getModule().isImdbUpdateAllowed(),
                                             getModule().isAmazonUpdateAllowed(),
                                             applyTemplate);
        itemForm.setVisible(true);
    }
    
    public void openUpdateAllDialog() {
        UpdateAllDialog dialog = new UpdateAllDialog(this);
        dialog.setVisible(true);
    }     
    

    public DcModule getModule() {
        return module;
    }     
    
    public abstract void applyGrouping(int[] fields);
    public abstract void undoGrouping();
    
    public abstract JComponent[] getExtraComponents();
    
    public abstract DcObject getObject(int row);
    
    public abstract Collection getSelectedObjects();
    public abstract DcObject getSelectedObject();
    
    public abstract DcObject getDcObject(String id);
    
    public abstract int[] getSelectedRows();
    public abstract int getRows();
    
    public abstract void applySettings();
    
    public abstract boolean isChangesSaved();
    
    public abstract void reload(String ID);
    
    public abstract void add(DcObject dco);
    public abstract void add(DcObject dco, boolean select);
    
    public abstract void add(Collection objects, boolean updateTree);
    public abstract void add(DcObject[] objects, boolean updateTree);

    public abstract void add(Collection objects);
    public abstract void add(DcObject[] objects);

    public abstract void clear();
    public abstract void clear(boolean safe);
    public abstract void open();
    public abstract void createReport();
    public abstract void delete();
    
    public abstract void removeRows(String[] ids);
    
    public abstract void save();
    public abstract void saveSelected();
    public abstract void updateRow(String ID, DcObject dco, boolean overwrite, boolean allowDeletes, boolean mark);
    public abstract void updateRow(int row, DcObject dco, boolean overwrite, boolean allowDeletes, boolean mark);
    
    public abstract void startAmazonUpdate();
    public abstract void startIMDBUpdate();
    
    public abstract void showQuickView(boolean b);
    public abstract void removeFromCache(String id);    
    
    public abstract void applyDefinitions();
}
