/******************************************************************************
 *                                     __                                     *
 *                              <-----/@@\----->                              *
 *                             <-< <  \\//  > >->                             *
 *                               <-<-\ __ /->->                               *
 *                               Data /  \ Crow                               *
 *                                   ^    ^                                   *
 *                       (c) 2003 The Data Crow team                          *
 *                              info@datacrow.net                             *
 *                                                                            *
 *                                                                            *
 *       This library is free software; you can redistribute it and/or        *
 *        modify it under the terms of the GNU Lesser General Public          *
 *       License as published by the Free Software Foundation; either         *
 *     version 2.1 of the License, or (at your option) any later version.     *
 *                                                                            *
 *      This library is distributed in the hope that it will be useful,       *
 *      but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU       *
 *           Lesser General Public License for more details.                  *
 *                                                                            *
 *     You should have received a copy of the GNU Lesser General Public       *
 *    License along with this library; if not, write to the Free Software     *
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA   *
 *                                                                            *
 ******************************************************************************/

package net.sf.dc.core.data;

import java.util.ArrayList;
import java.util.Collection;
import java.util.Collections;
import java.util.Iterator;

import net.sf.dc.core.modules.DcModules;
import net.sf.dc.core.objects.DcField;
import net.sf.dc.core.objects.DcMediaObject;
import net.sf.dc.core.objects.DcObject;
import net.sf.dc.core.objects.generic.GenericObject;
import net.sf.dc.resources.DcResources;
import net.sf.ext.util.CompositeComparator;
import net.sf.ext.util.StringUtils;

public class DataFilter {
    
    private String name = "";
    
    private int modIdx;
    private DcObject baseDco;
    private DataFilterOptions dfo;
    
    public DataFilter(int modIdx, DcObject baseDco, DataFilterOptions dfo) {
        this.modIdx = modIdx;
        this.baseDco = baseDco;
        this.dfo = dfo;
    }

    public DataFilter(String s) throws Exception {
        parse(s);
    }
    
    public void setBaseDco(DcObject baseDco) {
        this.baseDco = baseDco;
    }

    public DcField[] getOrdering() {
        return dfo != null ? dfo.getOrdering() : new DcField[0];
    }
    
    public boolean complyAll() {
        return dfo != null ? dfo.complyAll() : true;
    }

    public boolean exactMatch() {
        return dfo != null ? dfo.exactMatch() : false;
    }
    
    public DcObject getConditions() {
        return baseDco;
    }
    
    public String getName() {
        return name;
    }

    public void setName(String name) {
        this.name = name;
    }
    
    public int getModule() {
        return modIdx;
    }
    
    public boolean applies(DcObject dco) {
        boolean applies = applies(baseDco, dco);
        
        if (!applies && complyAll())
            return applies;
            
        dco.loadChildren();
        if (baseDco.getChildren().size() > 0) {
            for (Iterator itr = baseDco.getChildren().iterator(); itr.hasNext(); ) {
                DcObject baseChild = (DcObject) itr.next();
                
                if (baseChild.isChanged() && dco.getChildren().size() == 0) {
                    applies = false;
                    break;
                } else if (baseChild.isChanged() && dco.getChildren().size() > 0) {
                    for (Iterator iter = dco.getChildren().iterator(); iter.hasNext(); ) {
                        DcObject child = (DcObject) iter.next();
                        applies = applies(baseChild, child);
                        
                        if (applies)
                            break;
                    }
                }
            }
        }

        return applies;
    }
    
    private boolean applies(DcObject base, DcObject target) {
        int[] fields = target.getFieldIndices();
        String emptyCond = DcResources.getText("lblEmptySearchValue");
        
        if (base instanceof GenericObject){ 
        	DcObject o = target.getModule().getDcObject();
        	o.copy(base);
        	base = o;
        }
        
        boolean applies = true;
        for (int i = 0; i < fields.length; i++) {
            if ((base.getField(fields[i]).isUiOnly() || base.isChanged(fields[i]))) {
                
                DcField field = base.getField(fields[i]);
                
                boolean b = false;
                
                Object baseVal = base.getValue(fields[i]);
                Object targetVal = target.getValue(fields[i]);
                
                if (!base.isSearchable(fields[i]) && baseVal == null)
                    continue;
                
                if (field.isUiOnly() && baseVal == null)
                    continue;
                
                if (base.isChanged(fields[i])&& isEmpty(baseVal)) {

                    b = isEmpty(targetVal);
                
                } else if (!isEmpty(baseVal)) {
                    b = baseVal.toString().equals(emptyCond) ? isEmpty(targetVal) : !isEmpty(targetVal);
                        
                    if (b) {
                        if (dfo.exactMatch()) {
                            b = targetVal.toString().toLowerCase().equals(baseVal.toString().toLowerCase());
                        } else {
                            if (fields[i] == DcMediaObject._SYS_LOANDURATION) {
                                int duration1 = Integer.parseInt(baseVal.toString());
                                int duration2 = Integer.parseInt(targetVal.toString());
                                b = duration2 >= duration1;
                            } else {
                                b = targetVal.toString().toLowerCase().indexOf(baseVal.toString().toLowerCase()) > -1;
                            }
                        }
                    }
                } else {
                    b = true;
                }
                
                if (dfo.complyAll() && !b) {
                    applies = false;
                    break;
                } else if (dfo.complyAll()) {
                    applies = true;
                } else if (!dfo.complyAll() && !applies && b) {
                    applies = true;
                    break;
                } else {
                    applies = b;
                }
            }
        }        
        
        return applies;
    }
    
    private boolean isEmpty(Object o) {
        return o == null || o.toString().length() == 0 || o.toString().equals("-1");
    }    
    
    public void sort(Collection c) {
        if (dfo != null) {
            DcField[] fields = dfo.getOrdering();
            
            if (fields.length == 1)
                Collections.sort((ArrayList) c, new ResultSorter(fields[0]));
            else if (fields.length == 2)
                Collections.sort((ArrayList) c, 
                    new CompositeComparator(new ResultSorter(fields[0]), new ResultSorter(fields[1])));
        }
    }
    
    public String toString() {
        return getName();
    }
    
    public boolean equals(Object o) {
        String name1 = o instanceof DataFilter ? ((DataFilter) o).getName() : "";
        String name2 = getName();
        return name1.equals(name2);
    }
    
    private void parse(String s) throws Exception {
        this.modIdx = Integer.parseInt(StringUtils.getValueBetweenStrings("<MODULE>", "</MODULE>", s));
        this.baseDco = DcModules.get(modIdx).getDcObject();
        this.name = StringUtils.getValueBetweenStrings("<NAME>", "</NAME>", s);
        
        boolean exact = Boolean.getBoolean(StringUtils.getValueBetweenStrings("<EXACT>", "</EXACT>", s));
        boolean complyAll = Boolean.getBoolean(StringUtils.getValueBetweenStrings("<COMPLYALL>", "</COMPLYALL>", s));
        
        String fields = StringUtils.getValueBetweenStrings("<FIELDS>", "</FIELDS>", s);
        int idx = fields.indexOf("<FIELD>");
        DcObject child = baseDco.getModule().getChildModule() != null ? baseDco.getModule().getChildModule().getDcObject() : null;
        while (idx != -1) {
            String fieldDef = fields.substring(idx + 7, fields.indexOf("</FIELD>"));
            
            String table = StringUtils.getValueBetweenStrings("<TABLE>", "</TABLE>", fieldDef);
            String column = StringUtils.getValueBetweenStrings("<SYSNAME>", "</SYSNAME>", fieldDef);
            String value = StringUtils.getValueBetweenStrings("<VALUE>", "</VALUE>", fieldDef);
            
            DcObject dco = DcModules.getObjectForTable(table);
            if (dco.getModule().getIndex() == modIdx)
                baseDco.setValueForName(column, value);
            else if (child != null)
                child.setValueForName(column, value);
            
            fields = fields.substring(fields.indexOf("</FIELD>") + 8, fields.length());
            idx = fields.indexOf("<FIELD>");
        }
        
        String ordering = StringUtils.getValueBetweenStrings("<ORDERING>", "</ORDERING>", s);
        idx = ordering.indexOf("<FIELD>");
        Collection orderingFields = new ArrayList();
        while (idx != -1) {
            String column = ordering.substring(idx + 7, ordering.indexOf("</FIELD>"));
            
            for (Iterator iter = baseDco.getFields().iterator(); iter.hasNext(); ) {
                DcField field = (DcField) iter.next();
                if (field.getDatabaseFieldName().equals(column))
                    orderingFields.add(field);
            }
            
            ordering = ordering.substring(ordering.indexOf("</FIELD>") + 8, ordering.length());
            idx = ordering.indexOf("<FIELD>");
        }
        
        if (child != null)
            baseDco.addChild(child);
        
        DcField[] f = new DcField[orderingFields.size()];
        f = (DcField[]) orderingFields.toArray(f);
        dfo = new DataFilterOptions(f, complyAll, exact);
    }
    
    public String toStorageString() {
        String storage = "<FILTER>";
        
        storage += "<NAME>" + getName() + "</NAME>";
        storage += "<MODULE>" + getModule() + "</MODULE>";
        
        storage += "<OPTIONS>";
        storage += "<COMPLYALL>" + dfo.complyAll() + "</COMPLYALL>";
        storage += "<EXACT>" + dfo.exactMatch() + "</EXACT>";
        storage += "</OPTIONS>";
        
        storage += "<FIELDS>";
        storage += getFieldsString(baseDco);
        for (Iterator iter = baseDco.getChildren().iterator(); iter.hasNext(); ) {
            DcObject child = (DcObject) iter.next();
            storage += getFieldsString(child);
            break;
        }            
        
        storage += "</FIELDS>";

        storage += "<ORDERING>";
        if (dfo.getOrdering() != null) {
            for (int i = 0; i < dfo.getOrdering().length; i++) {
                storage += "<FIELD>" + dfo.getOrdering()[i].getDatabaseFieldName() + "</FIELD>";
            }             
        }
        storage += "</ORDERING>";
        
        storage += "</FILTER>";
        
        return storage;
    }
    
    private String getFieldsString(DcObject dco) {
        String s = "";
        for (Iterator iter = dco.getFields().iterator(); iter.hasNext(); ) {
            DcField field = (DcField) iter.next();
            Object value = dco.getValue(field.getIndex());
            if ((dco.isChanged(field.getIndex()) || field.isUiOnly()) && !isEmpty(value)) {
                s += "<FIELD>";
                s += "<TABLE>" + dco.getTableName() + "</TABLE>";
                s += "<SYSNAME>" + field.getSystemName() + "</SYSNAME>";   
                s += "<VALUE>" + (value instanceof DcObject ? ((DcObject) value).getID() : value) + "</VALUE>";
                s += "</FIELD>";    
            }
        } 
        return s;
    }    
}
