/******************************************************************************
 *                                     __                                     *
 *                              <-----/@@\----->                              *
 *                               <-<-\ __ /->->                               *
 *                               Data /  \ Crow                               *
 *                                   ^    ^                                   *
 *                       (c) 2003 The Data Crow team                          *
 *                              info@datacrow.net                             *
 *                                                                            *
 *                                                                            *
 *       This library is free software; you can redistribute it and/or        *
 *        modify it under the terms of the GNU Lesser General Public          *
 *       License as published by the Free Software Foundation; either         *
 *     version 2.1 of the License, or (at your option) any later version.     *
 *                                                                            *
 *      This library is distributed in the hope that it will be useful,       *
 *      but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU       *
 *           Lesser General Public License for more details.                  *
 *                                                                            *
 *     You should have received a copy of the GNU Lesser General Public       *
 *    License along with this library; if not, write to the Free Software     *
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA   *
 *                                                                            *
 ******************************************************************************/

package net.sf.dc.db;

import java.sql.Connection;
import java.sql.ResultSet;
import java.sql.ResultSetMetaData;
import java.sql.SQLException;
import java.sql.Statement;
import java.util.Collection;
import java.util.Iterator;

import net.sf.dc.core.Repository;
import net.sf.dc.core.modules.DcModule;
import net.sf.dc.core.modules.DcModules;
import net.sf.dc.core.objects.DcField;
import net.sf.dc.core.objects.DcObject;
import net.sf.dc.messages.Messages;
import net.sf.dc.resources.DcResources;
import net.sf.dc.settings.DcSettings;

public class DcDatabase {

    private QueryQueue queue;

    public DcDatabase()  {
    }

    public void initiliaze(Connection connection) throws Exception {
        startQueryQueue(connection);
        initialize(connection);
        setDbProperies(connection);

        new DatabaseConversion().start();

        try {
            connection.close();
        } catch (Exception exp) {
            Messages.add(exp);
        }
    }

    public int getQueueSize() {
    	return queue.getQueueSize();
    }
    
    private void startQueryQueue(Connection connection) {
        queue = new QueryQueue();
        Thread queryQueue = new Thread(queue, "queryQueue");
        queryQueue.setPriority(Thread.NORM_PRIORITY);
        queryQueue.setDaemon(true);
        queryQueue.start();
    }

    public String getName() {
        return DcSettings.getValueAsString(Repository.settings.stConnectionString);
    }

    public void addQuery(Query query) {
        queue.addQuery(query);
    }

    public void setDbProperies(Connection connection) {
        try {
            
            Statement stmt = connection.createStatement();
            int cacheScale = DcSettings.getValueAsInt(Repository.settings.stHsqlCacheScale);
            int cacheSizeScale = DcSettings.getValueAsInt(Repository.settings.stHsqlCacheSizeScale);

            stmt.execute("SET PROPERTY " +  '"' + "hsqldb.cache_scale" + '"' + " " + cacheScale);
            stmt.execute("SET PROPERTY " +  '"' + "hsqldb.cache_size_scale" + '"' + " " + cacheSizeScale);
            stmt.execute("SET LOGSIZE 200");
            
            stmt.close();
        } catch (Exception exp) {
            Messages.add(DcResources.getText("msgUnableToChangeDbSettings"), Messages._ERROR);
        }
    }

    private void initialize(Connection connection) throws Exception {
        Statement stmt = connection.createStatement();

        try {
            stmt.execute("SET DATABASE COLLATION " + '"' + "Latin1_General" + '"');
        } catch (Exception ignore) {}

        for (Iterator iter = DcModules.getAllModules();  iter.hasNext(); ) {
            DcModule module = (DcModule) iter.next();
            DcObject dco = module.getDcObject();

            if (!dco.getTableName().equals("")) {
                try {
                    String testQuery = "select * from " + dco.getTableName();
                    stmt.setMaxRows(1);
                    ResultSet result = stmt.executeQuery(testQuery);
                    initializeColumns(connection, result.getMetaData(), dco);
                    Messages.add(DcResources.getText("msgTableFound", dco.getTableName()), Messages._STANDARD);
                    result.close();
                } catch (SQLException exp) {
                    Messages.add(DcResources.getText("msgTableNotFound", dco.getTableName()), Messages._STANDARD);
                    createTable(connection, dco);
                }
            }
        }
        stmt.close();
    }

    private void initializeColumns(Connection connection, ResultSetMetaData metaData, DcObject dco) throws SQLException {
        Collection fields = dco.getFields();
        String tableName = dco.getTableName();

        int metaCount = metaData.getColumnCount();
        if (metaCount < dco.getDatabaseFieldCount()) {
            Messages.add(DcResources.getText("msgTableNeedsUpgrade", tableName), Messages._STANDARD);

            String[] metaColumns = new String[metaCount];
            for (int i = 1; i < metaCount + 1; i++) {
                metaColumns[i - 1] = metaData.getColumnName(i);
            }

            for (Iterator iter = fields.iterator(); iter.hasNext(); ) {
                DcField field = (DcField) iter.next();

                if (!field.isUiOnly()) {
                    String columnName = field.getDatabaseFieldName();
                    String columnType = field.getDataBaseFieldType();

                    boolean exists = false;
                    for (int j = 0; j < metaColumns.length; j++) {
                        if (columnName.equalsIgnoreCase(metaColumns[j]))
                            exists = true;
                    }

                    if (!exists) {
                        Statement stmt = connection.createStatement();
                        stmt.executeQuery("alter table " + tableName + " add column " + columnName + " " + columnType);
                        stmt.close();
                    }
                }
            }
        }
    }

    private void executeQuery(Connection connection, String query) {
        try {
            Statement stmt = connection.createStatement();
            stmt.executeQuery(query);
            Messages.add(query, Messages._QUERY);
        } catch (Exception ignore) {}
    }

    private void createTable(Connection connection, DcObject dco) {
        Query query = new Query(Query._CREATE, dco, null, null);
        executeQuery(connection, query.getQueries()[0]);

        try {
            DcObject[] objects = dco.getModule().getDefaultData();
            if (objects != null) {
                for (int i = 0; i < objects.length; i++) {
                    objects[i].setIDs();
                    objects[i].keepInSync(false);
                    objects[i].setSilence(true);
                    objects[i].saveNew(false);
                }
            }
        } catch (Exception exp) {
            Messages.add(exp);
        }
    }
}
