/******************************************************************************
 *                                     __                                     *
 *                              <-----/@@\----->                              *
 *                             <-< <  \\//  > >->                             *
 *                               <-<-\ __ /->->                               *
 *                               Data /  \ Crow                               *
 *                                   ^    ^                                   *
 *                       (c) 2003 The Data Crow team                          *
 *                              info@datacrow.net                             *
 *                                                                            *
 *                                                                            *
 *       This library is free software; you can redistribute it and/or        *
 *        modify it under the terms of the GNU Lesser General Public          *
 *       License as published by the Free Software Foundation; either         *
 *     version 2.1 of the License, or (at your option) any later version.     *
 *                                                                            *
 *      This library is distributed in the hope that it will be useful,       *
 *      but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU       *
 *           Lesser General Public License for more details.                  *
 *                                                                            *
 *     You should have received a copy of the GNU Lesser General Public       *
 *    License along with this library; if not, write to the Free Software     *
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA   *
 *                                                                            *
 ******************************************************************************/

package net.sf.dc.db;

import java.text.SimpleDateFormat;
import java.util.ArrayList;
import java.util.Collection;
import java.util.Date;
import java.util.Iterator;

import javax.swing.ImageIcon;

import net.sf.dc.core.Repository;
import net.sf.dc.core.data.DataManager;
import net.sf.dc.core.modules.DcMediaModule;
import net.sf.dc.core.modules.DcModule;
import net.sf.dc.core.modules.DcModules;
import net.sf.dc.core.objects.DcField;
import net.sf.dc.core.objects.DcMapping;
import net.sf.dc.core.objects.DcObject;
import net.sf.dc.core.objects.Loan;
import net.sf.dc.core.objects.Picture;
import net.sf.dc.core.objects.audiocd.AudioCD;
import net.sf.dc.core.objects.audiocd.AudioTrack;
import net.sf.dc.core.objects.contactperson.ContactPerson;
import net.sf.dc.core.objects.musicalbum.MusicAlbum;
import net.sf.dc.core.objects.musicalbum.MusicTrack;
import net.sf.dc.resources.DcResources;
import net.sf.dc.wf.requests.ImageRequest;
import net.sf.dc.wf.requests.RequestCollection;

public class Query {

    public static final int _UNDEFINED = 0;
    public static final int _UPDATE = 1;
    public static final int _CREATE = 2;
    public static final int _INSERT = 3;
    public static final int _SELECT = 4;
    public static final int _DELETE = 5;

    private static final int _SELECTPRECISE = 6;
    private static final int _SELECTOR = 7;
    private static final int _SELECTPRECISEOR = 8;

    private final SimpleDateFormat formatter = new SimpleDateFormat("yyyy-MM-dd");

    private final int module;

    private RequestCollection requestors;
    private int queryType;

    private boolean loadRelations = true;
    private boolean bPreciseSelect = false;
    private boolean bComplyToAllConditions = true;

    private String[] queries;
    private String objectID;

    private String[] ordering;
    private boolean bSilence = false;
    private boolean isBatch = false;
    private boolean endOfBatch = false;

    public Query(   int queryType,
                    DcObject dco,
                    QueryOptions options,
                    RequestCollection requestors) {

        module = dco.getModule().getIndex();

        setSilence(dco.getSilence());

        if (options != null) {
            bPreciseSelect = options.getPreciseSelect();
            bComplyToAllConditions = options.getComplyAllConditions();
            ordering = options.getOrdering();
        }

        this.requestors = requestors;
        this.requestors = requestors == null ? new RequestCollection() : this.requestors;
        this.queryType = queryType;
        this.objectID = dco.getID();

        if (this.queryType == _SELECT) {
            if (bPreciseSelect && !bComplyToAllConditions) {
                this.queryType = _SELECTPRECISEOR;
            } else if (bPreciseSelect) {
                this.queryType = _SELECTPRECISE;
            } else if (!bComplyToAllConditions) {
                this.queryType = _SELECTOR;
            }
        }

        queries = getQueries(dco);
    }

    public void unload() {
        if (requestors != null)
            requestors.clear();
        
        requestors = null;
        queries = null;
        objectID = null;
        ordering = null;
    }
    
    public int getModule() {
        return module;
    }

    public String[] getOrdering() {
        return ordering;
    }

    public void setLoadRelations(boolean b) {
        loadRelations = b;
    }

    public boolean getLoadRelations() {
        return loadRelations;
    }

    public void setBatch(boolean endOfBatch) {
        this.isBatch = true;
        this.endOfBatch = endOfBatch;
    }

    public boolean isBatch() {
        return isBatch;
    }

    public boolean isEndOfBatch() {
        return isBatch() && endOfBatch;
    }

    public String getObjectID() {
        return objectID;
    }

    public int getQueryType() {
        return queryType;
    }

    public boolean getSilence() {
        return bSilence;
    }

    public void setSilence(boolean silence) {
        this.bSilence = silence;
    }

    public RequestCollection getRequestors() {
        return requestors;
    }

    public String[] getQueries() {
        return queries;
    }

    private String[] getQueries(DcObject dco) {
        String[] queries = new String[1];
        switch (queryType) {
            case _UPDATE :
                queries = getUpdateQueries(dco);
                break;
            case _CREATE :
                queries[0] = getCreateTableQuery(dco);
                break;
            case _INSERT :
                queries = getInsertQueries(dco);
                break;
            case _DELETE :
                queries = getDeleteQueries(dco);
                break;
            case _UNDEFINED :
            case _SELECT :
            case _SELECTPRECISE :
            case _SELECTOR :
            case _SELECTPRECISEOR :
                queries = getSelectQueries(dco);
                break;
        }

        return queries;
    }

    private boolean isEmpty(String s) {
        String emptyValue = DcResources.getText("lblEmptySearchValue");
        return s.equals("''") || s.equals("'" + emptyValue + "'") ||
               s.equals(emptyValue) || s.equals("0") || s.equals("NULL");
    }

    private String getEmptyCondition(String column, String s, DcField field) {
        String emptyCondition;
        if (isInteger(field)) 
            emptyCondition = "(" + column + " = 0 OR " + column + " IS NULL) ";
        else
            emptyCondition = "(" + column + " IS NULL OR " + column + " = '') ";

        return emptyCondition;
    }

    private String getQueryValue(DcObject object, int index) {
        Object value = object.getValue(index);

        String queryValue = "";
        if (value == null || value.toString().trim().length() == 0)
            queryValue = "NULL";
        else if (value instanceof DcObject)
            queryValue = ((DcObject) value).getID();
        else if (value instanceof Date && value != null)
            queryValue = formatter.format((Date) value);
        else
            queryValue = value.toString();

        boolean isInteger = isInteger(object.getField(index));
        boolean isBoolean = isBoolean(object.getField(index));

        if (value != null && value.toString().length() > 0 && !queryValue.equals("NULL"))
            queryValue = !isInteger && !isBoolean ? "'" + queryValue + "'" : queryValue;

        return queryValue;
    }
    
    private boolean isBoolean(DcField field) {
        return  field.getValueType() == Repository.datatypes._BOOLEAN;
    }
    
    private boolean isInteger(DcField field) {
        return  field.getValueType() == Repository.datatypes._INTEGER ||
                field.getValueType() == Repository.datatypes._BIGINTEGER || 
                field.getValueType() == Repository.datatypes._DCOBJECTREFERENCE ||
                field.getValueType() == Repository.datatypes._DCPARENTREFERENCE;
    }

    private String[] getInsertQueries(DcObject dco) {
        StringBuffer values = new StringBuffer();
        StringBuffer columns = new StringBuffer();

        Collection references = new ArrayList();
        Collection pictures = new ArrayList();
        Collection fields = dco.getFields();
        for (Iterator iter = fields.iterator(); iter.hasNext(); ) {
            DcField field = (DcField) iter.next();

            if (field.getValueType() == Repository.datatypes._PICTURE) {
                Picture picture = (Picture) dco.getValue(field.getIndex());
                byte[] image = picture != null && picture.isNew() ? (byte[]) picture.getValue(Picture._D_IMAGE) : null;
                if (image != null && image.length > 10) {
                    ImageIcon icon =  new ImageIcon(image);
                    picture.setValue(Picture._A_OBJECTID, dco.getID());
                    picture.setValue(Picture._B_FIELD, field.getDatabaseFieldName());
                    picture.setValue(Picture._C_FILENAME, dco.getID() + "_" + field.getDatabaseFieldName() + ".jpg");
                    picture.setValue(Picture._E_HEIGHT, new Integer(icon.getIconHeight()));
                    picture.setValue(Picture._F_WIDTH, new Integer(icon.getIconWidth()));
                    pictures.add(picture);
                }
            } else if (field.getValueType() == Repository.datatypes._DCOBJECTCOLLECTION) {
                Collection c = (Collection) dco.getValue(field.getIndex());
                if (c != null) references.addAll(c);                
                
            } else if (dco.isChanged(field.getIndex()) && !field.isUiOnly()) {
                if (values.length() > 0)
                    values.append(", ");

                if (columns.length() > 0)
                    columns.append(", ");

                if (!(dco.getValue(field.getIndex()) instanceof Picture)) {
                    columns.append(field.getDatabaseFieldName());
                    values.append(getQueryValue(dco, field.getIndex()));
                }
            }
        }

        int index = 0;
        String[] queries = new String[(pictures.size() + 1) + (references.size() + 1)];
        
        if (references.size() > 0) {
            for (Iterator iter = references.iterator(); iter.hasNext(); ) {
                DcMapping mapping = (DcMapping) iter.next();
                queries[index++] = "INSERT INTO " + mapping.getTableName() + 
                                    " (" + mapping.getDatabaseFieldName(DcMapping._A_MEDIA_OBJECT_ID) + ", " +
                                           mapping.getDatabaseFieldName(DcMapping._B_SIMPLE_OBJECT_ID) + 
                                      ") \r\n VALUES (" + dco.getID() + ", " + mapping.getSimpleObjectId() + ");";
            }
        }        
        
        queries[index++] = "INSERT INTO " + dco.getTableName() + " (" + columns + ") \r\n" + "VALUES (" + values + ");";
        for (Iterator iter = pictures.iterator(); iter.hasNext(); ) {
            Picture picture = (Picture) iter.next();
            queries[index++] = getInsertQueries(picture)[0];
            requestors.addRequest(
                    new ImageRequest((byte[]) picture.getValue(Picture._D_IMAGE), 
                                     (String) picture.getValue(Picture._C_FILENAME), ImageRequest._SAVE));
        }
        
        pictures.clear();
        return queries;
    }

    private String getCreateTableQuery(DcObject dco) {
        String columns = "";
        Collection fields = dco.getFields();
        for (Iterator iter = fields.iterator(); iter.hasNext(); ) {
            DcField field = (DcField) iter.next();

            if (!field.isUiOnly()) {
                if (columns.length() > 0)
                    columns += ", ";

                columns += field.getDatabaseFieldName() + " " + field.getDataBaseFieldType();
                if (field.getIndex() == DcObject._ID) {
                    columns += " PRIMARY KEY";
                }
            }
        }
        return "CREATE MEMORY TABLE " + dco.getTableName() + "\r\n(" + columns + ");";
    }

    private String[] getUpdateQueries(DcObject dco) {
        Collection pictures = new ArrayList();
        Collection references = new ArrayList();
        
        StringBuffer values = new StringBuffer();
        Collection fields = dco.getFields();
        for (Iterator iter = fields.iterator(); iter.hasNext(); ) {
            DcField field = (DcField) iter.next();

            if (field.getValueType() == Repository.datatypes._PICTURE) {
                Picture picture = (Picture) dco.getValue(field.getIndex());
                if (picture != null && (picture.isNew() || picture.isUpdated() || picture.isDeleted())) {
                    picture.setValue(Picture._A_OBJECTID, dco.getID());
                    picture.setValue(Picture._B_FIELD, field.getDatabaseFieldName());
                    picture.setValue(Picture._C_FILENAME, dco.getID() + "_" + field.getDatabaseFieldName() + ".jpg");
                    pictures.add(picture);
                    
                    if (picture.getValue(Picture._D_IMAGE) != null) {
                        byte[] image = (byte[]) picture.getValue(Picture._D_IMAGE);
                        ImageIcon icon = new ImageIcon(image);
                        picture.setValue(Picture._E_HEIGHT, new Integer(icon.getIconHeight()));
                        picture.setValue(Picture._F_WIDTH, new Integer(icon.getIconWidth()));
                    }
                    
                    picture.setValue(Picture._A_OBJECTID, dco.getID());
                    picture.setValue(Picture._B_FIELD, field.getDatabaseFieldName());
                    picture.setValue(Picture._C_FILENAME, dco.getID() + "_" + field.getDatabaseFieldName() + ".jpg");
                }
            } else if (field.getValueType() == Repository.datatypes._DCOBJECTCOLLECTION) {
                Collection c = (Collection) dco.getValue(field.getIndex());
                if (c != null) references.addAll(c);
                
            } else if (dco.isChanged(field.getIndex()) && !field.isUiOnly()) {
                if (values.length() > 0)
                    values.append(", ");

                values.append("\r\n");
                values.append(field.getDatabaseFieldName());
                values.append(" = ");

                values.append(getQueryValue(dco, field.getIndex()));
            }
        }

        int index = 0;
        String[] queries = new String[(pictures.size() + 1) + (references.size() + 1)];

        if (!(dco instanceof Picture))
            queries[index++] = "UPDATE " + dco.getTableName() + " SET " + values + "\r\n WHERE ID = " + dco.getID() + ";";
        else
            queries[index++] = "UPDATE " + dco.getTableName() + " SET " + values + "\r\n WHERE " +
                                dco.getDatabaseFieldName(Picture._A_OBJECTID) + " = " + dco.getValue(Picture._A_OBJECTID) + " AND " +
                                dco.getDatabaseFieldName(Picture._B_FIELD) + " = '" + dco.getValue(Picture._B_FIELD) + "';";

        if (references.size() > 0) {
            boolean first = true;
            for (Iterator iter = references.iterator(); iter.hasNext(); ) {
                DcMapping mapping = (DcMapping) iter.next();
                if (first)
                    queries[index++] = "DELETE FROM " + mapping.getTableName() + " WHERE " +  
                                        mapping.getDatabaseFieldName(DcMapping._A_MEDIA_OBJECT_ID) + " = " + dco.getID();

                queries[index++] = "INSERT INTO " + mapping.getTableName() + 
                                    " (" + mapping.getDatabaseFieldName(DcMapping._A_MEDIA_OBJECT_ID) + ", " +
                                           mapping.getDatabaseFieldName(DcMapping._B_SIMPLE_OBJECT_ID) + 
                                      ") \r\n VALUES (" + dco.getID() + ", " + mapping.getSimpleObjectId() + ");";
                first = false;
            }
        }
        
        for (Iterator iter = pictures.iterator(); iter.hasNext(); ) {
            Picture picture = (Picture) iter.next();
            
            Collection currentPictures = DataManager.getPictures(dco.getID());
            boolean isReallyNew = picture.isNew() && (currentPictures == null || !currentPictures.contains(picture));
            if (!isReallyNew && picture.isNew())
                picture.isUpdated(!isReallyNew);
            
            if (picture.isUpdated()) {
                queries[index++] = getUpdateQueries(picture)[0];
                requestors.addRequest(
                        new ImageRequest((byte[]) picture.getValue(Picture._D_IMAGE), 
                                         (String) picture.getValue(Picture._C_FILENAME), ImageRequest._SAVE));                
            } else if (picture.isDeleted()) {
                queries[index++] = "DELETE FROM " + picture.getTableName() + " WHERE " +
                                   picture.getField(Picture._A_OBJECTID).getDatabaseFieldName() + " = " + dco.getID() + " AND " +
                                   picture.getField(Picture._B_FIELD).getDatabaseFieldName() + " = '" + picture.getValue(Picture._B_FIELD) + "';";
                requestors.addRequest(
                        new ImageRequest((byte[]) picture.getValue(Picture._D_IMAGE), 
                                         (String) picture.getValue(Picture._C_FILENAME), ImageRequest._DELETE));                
            } else if (picture.isNew()) {
                queries[index++] = getInsertQueries(picture)[0];
                requestors.addRequest(
                        new ImageRequest((byte[]) picture.getValue(Picture._D_IMAGE), 
                                         (String) picture.getValue(Picture._C_FILENAME), ImageRequest._SAVE));
            }
        }

        pictures.clear();
        return queries;
    }

    private String[] getDeleteQueries(DcObject dco) {
        Collection queries = new ArrayList();

        Loan loan = new Loan();
        Picture picture = new Picture();
        if (dco.hasPrimaryKey())
            queries.add("DELETE FROM " + dco.getTableName() + " WHERE ID = " + dco.getID());

        if (dco.getModule().canBeLended())
            queries.add("DELETE FROM " + loan.getTableName() + " WHERE " +
                        loan.getField(Loan._D_OBJECTID).getDatabaseFieldName() + " = " + dco.getID());

        if (dco instanceof AudioCD) 
            queries.add("DELETE FROM " + new AudioTrack().getTableName() + " WHERE ALBUM  = " + dco.getID());

        if (dco instanceof MusicAlbum) 
            queries.add("DELETE FROM " + new MusicTrack().getTableName() + " WHERE ALBUM  = " + dco.getID());
        
        requestors.addRequest(new ImageRequest(dco.getID(), ImageRequest._DELETE));
        
        queries.add("DELETE FROM " + picture.getTableName() + " WHERE " +
                    picture.getField(Picture._A_OBJECTID).getDatabaseFieldName() + " = " + dco.getID());

        String[] tmp = new String[queries.size()];
        return (String[]) queries.toArray(tmp);
    }

    private void addAvailabilityCondition(DcObject dco, StringBuffer conditions) {
        if (dco.getModule().canBeLended()) {
            ContactPerson loanedBy = (ContactPerson) dco.getValue(DcObject._SYS_LOANEDBY);
            Integer duration = (Integer) dco.getValue(DcObject._SYS_LOANDURATION);
            String s = (String) dco.getValue(DcObject._SYS_AVAILABLE);

            if (s != null || loanedBy != null || duration != null) {
                boolean available = s == null || duration != null ? false : new Boolean(s).booleanValue();
                loanedBy = available ? null : loanedBy;

                if (conditions.length() > 0)
                    conditions.append(" AND");

                String column = dco.getChildren().size() > 0 ? " " + dco.getTableShortName() + ".ID" : " ID";
                String tablename = dco.getChildren().size() > 0 ? dco.getTableShortName() : dco.getTableName();

                String current = formatter.format(new Date());
                String daysCondition = duration != null ? " AND DATEDIFF('dd', startDate , '" + current + "') >= " + duration.intValue() : "";
                String personCondition = loanedBy != null ? " AND PersonID = " + loanedBy.getID() : "";

                if (available)
                    conditions.append(column + " NOT in (select objectID from Loans where objectID = " +
                                      tablename + ".ID AND enddate IS NULL AND startDate <= '" + current +  "')");
                else
                    conditions.append(column + " in (select objectID from Loans where objectID = " +
                                      tablename + ".ID " + daysCondition + " AND enddate IS NULL AND startDate <= '" + current +  "'" +
                                      personCondition + ")");
            }
        }
    }

    private String[] getSelectQueries(DcObject dco) {
        String[] queries;
        int tableCount = 0;

        Collection objects = new ArrayList();

        Collection tables = new ArrayList();
        if (dco.getModule().getIndex() == DcModules._GENERIC) {
            for (Iterator iter = DcModules.getModules(); iter.hasNext(); ) {
                DcModule module = (DcModule) iter.next();
                
                if (    module instanceof DcMediaModule && module.getIndex() != DcModules._GENERIC &&
                        module.getIndex() != DcModules._AUDIOTRACK &&
                        module.getIndex() != DcModules._MUSICTRACK) {
                    
                    tables.add(module.getTableName());
                    tableCount++;
                }
            }
        } else {
            objects.add(dco);
            objects.addAll(dco.getChildren());

            tables.add(dco.getTableJoin());
            tableCount++;
        }

        queries = new String[tableCount];

        int counterTables = 0;
        for (Iterator iterTables = tables.iterator(); iterTables.hasNext(); counterTables++) {
            StringBuffer columns = new StringBuffer();

            String query = "";
            String table = (String) iterTables.next();

            if (dco.getModule().getIndex() == DcModules._GENERIC) {
                objects.clear();
                DcObject o = DcModules.getObjectForTable(table);
                o.copy(dco);
                objects.add(o);
            }

            int counter = 0;
            StringBuffer conditions = new StringBuffer();

            for (Iterator iter = objects.iterator(); iter.hasNext(); counter++) {
                DcObject object = (DcObject) iter.next();
                Collection fields = object.getFields();

                for (Iterator iterFields = fields.iterator(); iterFields.hasNext(); ) {
                    DcField field = (DcField) iterFields.next();

                    if (!field.isUiOnly() && field.getValueType() != Repository.datatypes._PICTURE)  {
                        String column = field.getDatabaseFieldName();
                        if (objects.size() > 1)
                            column = object.getTableShortName() + "." + column;

                        if (counter == 0) {
                            if (columns.length() > 0)
                                columns.append(", ");

                            // only select fields for the current module, child fields are added to the condition only
                            columns.append(column);
                        }

                        if (object.isChanged(field.getIndex())) {
                            String value = getQueryValue(object, field.getIndex());
                            String condition = "";

                            String appender = queryType == _SELECTOR || queryType == _SELECTPRECISEOR ? " OR " : " AND ";
                            if (conditions.length() > 0)
                                conditions.append(appender);
                            
                            if (isEmpty(value)) {
                                condition = getEmptyCondition(column, value, field);
                            } else {
                                if (queryType == _SELECT || queryType == _SELECTOR) {
                                    if (value.startsWith("'"))
                                        condition = "UPPER (" + column + ") LIKE '%" + value.substring(1, value.length() -1).toUpperCase() + "%'";
                                    else 
                                        condition = column + " = " + value;
                                } else {
                                    
                                    if (value.startsWith("'")) 
                                        condition = "UPPER (" + column + ") = " + value.toUpperCase();
                                    else
                                        condition = column + " = " + value;
                                }
                            }
                            conditions.append(condition);
                        }
                    }
                }
            }

//            int max = DcSettings.getValueAsInt(Repository.settings.stMaximumSearchRes);
//            if (dco instanceof Picture)
//                max = 1;
            
            query+= "SELECT DISTINCT " + columns + " FROM " + table;

            DcObject objectClone = DcModules.getObjectForTable(table);
            if (objectClone  != null)
                objectClone.copy(dco);

            addAvailabilityCondition(objectClone == null ? dco : objectClone, conditions);

            if (conditions.length() > 0)
                query +=  "\r\n WHERE " + conditions;

            if (ordering != null) {
                for (int i = 0; i < ordering.length; i++) {
                    if (i == 0)
                        query += " ORDER BY ";
                    else
                        query += ", ";

                    query += ordering[i];
                }
            } else if (dco.hasPrimaryKey()) {
                query += " ORDER BY " + dco.getField(DcObject._ID).getDatabaseFieldName();
            }

            queries[counterTables] = query + ";";
        }

        return queries;
    }

    public void finalize() throws Throwable {
        try {
            queries = null;
            objectID = null;
            requestors = null;
            ordering = null;
        } finally {
            super.finalize();
        }
    }
}
