/******************************************************************************
 *                                     __                                     *
 *                              <-----/@@\----->                              *
 *                             <-< <  \\//  > >->                             *
 *                               <-<-\ __ /->->                               *
 *                               Data /  \ Crow                               *
 *                                   ^    ^                                   *
 *                       (c) 2003 The Data Crow team                          *
 *                              info@datacrow.net                             *
 *                                                                            *
 *                                                                            *
 *       This library is free software; you can redistribute it and/or        *
 *        modify it under the terms of the GNU Lesser General Public          *
 *       License as published by the Free Software Foundation; either         *
 *     version 2.1 of the License, or (at your option) any later version.     *
 *                                                                            *
 *      This library is distributed in the hope that it will be useful,       *
 *      but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU       *
 *           Lesser General Public License for more details.                  *
 *                                                                            *
 *     You should have received a copy of the GNU Lesser General Public       *
 *    License along with this library; if not, write to the Free Software     *
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA   *
 *                                                                            *
 ******************************************************************************/

package net.sf.dc.db;

import java.sql.Connection;
import java.sql.ResultSet;
import java.sql.SQLException;
import java.sql.Statement;
import java.util.ArrayList;
import java.util.Collection;
import java.util.Iterator;
import java.util.LinkedList;

import net.sf.dc.console.dialogs.MessageBox;
import net.sf.dc.console.views.MasterView;
import net.sf.dc.core.modules.DcModules;
import net.sf.dc.core.objects.DcObject;
import net.sf.dc.messages.Messages;
import net.sf.dc.resources.DcResources;
import net.sf.dc.wf.WorkFlow;
import net.sf.dc.wf.requests.IRequest;
import net.sf.dc.wf.requests.IUpdateUIRequest;
import net.sf.dc.wf.requests.ImageRequest;
import net.sf.dc.wf.requests.RequestCollection;

public class QueryQueue extends Thread {

    private final LinkedList lQueryQueue = new LinkedList();

    private boolean isLazy = true;
    private boolean success = true;
    
    private boolean debug = false;

    public QueryQueue() {
    }

    public int getQueueSize() {
    	return lQueryQueue.size();
    }
    
    public void addQuery(Query query) {
        lQueryQueue.addLast(query);
    }

    public void run() {
        synchronized(this) {
            while (true) {
                try {
                    if (lQueryQueue.size() > 0 && isLazy) {
                        success = true;
                        isLazy = false;

                        Query query = (Query) lQueryQueue.removeFirst();
                        Collection result = executeQuery(query);
                    	handleRequests(query, result, success);
                        query.unload();
                    } else {
                        isLazy = true;
                        sleep(100);
                    }
                } catch (Exception exp) {
                    Messages.add(exp);
                }
            }
        }
    }

    private Collection executeQuery(Query query) {
        switch(query.getQueryType()) {
            case Query._UPDATE:
                success = executeUpdateQuery(query);
                break;
            case Query._DELETE:
                success = executeDeleteQuery(query);
                break;
            case Query._INSERT:
                success = executeInsertQuery(query);
                break;
            case Query._UNDEFINED:
                if (!query.getSilence()) {
                	Messages.add(query.getQueries()[0], Messages._QUERY);
                }
                return executeUndefinedQuery(query);
            default:
                if (!query.getSilence() || debug) {
                    Messages.add(query.getQueries()[0], Messages._QUERY);
                }
                return executeSelectQuery(query);
        }

        if (debug) 
        	Messages.add(query.getQueries()[0], Messages._QUERY);

        return null;
    }

    private boolean executeUpdateQuery(Query query) {
        String[] queries = query.getQueries();

        Statement stmt = null;
        Connection connection = DatabaseManager.getConnection();

        boolean success = true;
        boolean messageShown = false;
        for (int i = 0; i < queries.length && success; i++) {
            String sQuery = queries[i];
            
            if (sQuery != null) {
                try {
                    stmt = connection.createStatement();
                    stmt.executeUpdate(sQuery);
                    if ((query.isBatch() && query.isEndOfBatch() && !query.getSilence() && !messageShown) ||
                        (!query.getSilence() && !messageShown)) {
                        messageShown = true;
                        new MessageBox(DcResources.getText("msgDataSaved"), MessageBox._INFORMATION, false);
                    }
                    stmt.close();
                } catch (SQLException sqlExp) {
                    success = false;
                    Messages.add(sQuery, Messages._QUERY);
                    Messages.add(sqlExp);
                    new MessageBox(DcResources.getText("msgQueryError", sqlExp.toString()), MessageBox._ERROR, false);
                }
            }
        }

        closeConnection(connection, stmt);
        return success;
    }

    private boolean executeInsertQuery(Query query) {
        String[] queries = query.getQueries();

        Statement stmt = null;
        Connection connection = DatabaseManager.getConnection();

        boolean success = true;
        boolean messageShown = false;
        for (int i = 0; i < queries.length && success; i++) {
            String qry = queries[i];
            
            if (qry != null) {
                try {
                    stmt = connection.createStatement();
                    stmt.execute(qry);
                    if ((query.isBatch() && query.isEndOfBatch() && !query.getSilence() && !messageShown) ||
                            (!query.getSilence() && !messageShown)) {
                        messageShown = true;
                        new MessageBox(DcResources.getText("msgDataSaved"), MessageBox._INFORMATION, false);
                    }

                    stmt.close();
                } catch (SQLException sqlExp) {
                    success = false;
                    Messages.add(qry, Messages._QUERY);
                    Messages.add(sqlExp);
                    //new MessageBox(DcResources.getText("msgQueryError", sqlExp.toString()), MessageBox._ERROR, false);
                }
            }
        }

        closeConnection(connection, stmt);
        return success;
    }

    private boolean executeDeleteQuery(Query query) {
        String[] queries = query.getQueries();

        Statement stmt = null;
        Connection connection = DatabaseManager.getConnection();

        boolean messageShown = false;
        boolean success = true;
        for (int i = 0; i < queries.length; i++) {
            String sQuery = queries[i];
            
            try {
                stmt = connection.createStatement();
                stmt.executeQuery(sQuery);
                if (    (query.isBatch() && query.isEndOfBatch() && !query.getSilence() && !messageShown) ||
                        (!query.getSilence() && !messageShown)) {
                    messageShown = true;
                    new MessageBox(DcResources.getText("msgDeleteSuccessfull"), MessageBox._INFORMATION, false);
                }

                stmt.close();
            } catch (SQLException sqlExp) {
                success = false;
                Messages.add(sQuery, Messages._QUERY);
                Messages.add(sqlExp);
                new MessageBox(DcResources.getText("msgQueryError", sqlExp.toString()), MessageBox._ERROR, false);
            }
        }

        closeConnection(connection, stmt);

        return success;
    }

    private Collection executeSelectQuery(Query query) {
        String[] queries = query.getQueries();
        ResultSet result = null;

        Collection objects = new ArrayList();
        for (int i = 0; i < queries.length; i++) {
            String sQuery = queries[i];
            
            Statement stmt = null;
            Connection connection = DatabaseManager.getConnection();
            try {
                stmt = connection.createStatement();
                result = stmt.executeQuery(sQuery);
            } catch (SQLException sqlExp) {
                Messages.add(sQuery, Messages._QUERY);
                Messages.add(sqlExp);

                if (sqlExp.getMessage().toUpperCase().indexOf("MEMORY") > -1) {
                    new MessageBox(DcResources.getText("msqMemoryError"), MessageBox._ERROR, false);
                    System.gc();
                } else {
                    new MessageBox(DcResources.getText("msgQueryError", sqlExp.toString()), MessageBox._ERROR, false);
                }
            }

            objects.addAll(new WorkFlow().convertToDCObjects(result));

            if (query.getLoadRelations()) {
                for (Iterator iter = objects.iterator(); iter.hasNext(); ) {
                    ((DcObject) iter.next()).loadChildren();
                }
            }

            try {
                result.close();
                closeConnection(connection, stmt);
            } catch (SQLException exp) {
                Messages.add(exp);
            }
        }

        return objects;
    }
    
    private Collection executeUndefinedQuery(Query query) {
        String sQuery = query.getQueries()[0];
        ResultSet result = null;

        Statement stmt = null;
        Connection connection = DatabaseManager.getConnection();

        try {
            stmt = connection.createStatement();
            result = stmt.executeQuery(sQuery);
        } catch (SQLException sqlExp) {
            Messages.add(sQuery, Messages._QUERY);
            Messages.add(sqlExp);

            if (sqlExp.getMessage().toUpperCase().indexOf("MEMORY") > -1) {
                new MessageBox(DcResources.getText("msqMemoryError"),
                               MessageBox._ERROR, false);
                System.gc();
            } else {
                new MessageBox(DcResources.getText("msgQueryError", sqlExp.toString()),
                               MessageBox._ERROR, false);
            }
        }

        Collection objects = new WorkFlow().convertToDCObjects(result);
        try {
            result.close();
            closeConnection(connection, stmt);
        } catch (SQLException exp) {
            Messages.add(exp);
        }
        return objects;
    }

    private void handleRequests(Query qry,
                                Collection objects,
                                boolean qryWasSuccess) {
        
        RequestCollection requestors = qry.getRequestors();
        if (requestors != null)  {
            RequestCollection uiRequests = new RequestCollection();
            RequestCollection imageRequests = new RequestCollection();
            
            IRequest[] requests = requestors.getRequests();
            for (int i = 0; i < requests.length; i++) {
                requestors.removeRequestor(requests[i]);
                
                if (requests[i] instanceof ImageRequest)
                    imageRequests.addRequest(requests[i]);
                else if (requests[i] instanceof IUpdateUIRequest)
                    uiRequests.addRequest(requests[i]);
                else
                    requests[i].execute(objects);
            }
            
            boolean hasImageRequest = imageRequests.size() > 0;
            IRequest[] imgRequests = imageRequests.getRequests();
            for (int i = 0; i < imgRequests.length; i++) {
                imageRequests.removeRequestor(imgRequests[i]);
                imgRequests[i].execute(objects);
            }
            
            if (hasImageRequest && qry.getObjectID() != null) {
                MasterView[] views = DcModules.get(qry.getModule()).getViews();
                for (int i = 0; i < views.length; i++)
                    views[i].reload(qry.getObjectID());
            }
            
            WorkFlow.handleRequests(objects, uiRequests, qryWasSuccess);
        }
    }

    private void closeConnection(Connection connection, Statement stmt) {
        try {
            connection.commit();
            connection.close();

            if (stmt != null)
                stmt.close();
        } catch (SQLException exp) {
            Messages.add(exp);
        }
    }
}
