/******************************************************************************
 *                                     __                                     *
 *                              <-----/@@\----->                              *
 *                             <-< <  \\//  > >->                             *
 *                               <-<-\ __ /->->                               *
 *                               Data /  \ Crow                               *
 *                                   ^    ^                                   *
 *                       (c) 2003 The Data Crow team                          *
 *                              info@datacrow.net                             *
 *                                                                            *
 *                                                                            *
 *       This library is free software; you can redistribute it and/or        *
 *        modify it under the terms of the GNU Lesser General Public          *
 *       License as published by the Free Software Foundation; either         *
 *     version 2.1 of the License, or (at your option) any later version.     *
 *                                                                            *
 *      This library is distributed in the hope that it will be useful,       *
 *      but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU       *
 *           Lesser General Public License for more details.                  *
 *                                                                            *
 *     You should have received a copy of the GNU Lesser General Public       *
 *    License along with this library; if not, write to the Free Software     *
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA   *
 *                                                                            *
 ******************************************************************************/

package net.sf.dc.processes.cdimport;

import java.io.File;
import java.util.Collection;
import java.util.Iterator;
import java.util.Vector;

import net.sf.dc.console.dialogs.MessageBox;
import net.sf.dc.console.importcd.ImportCDDialog;
import net.sf.dc.core.DataCrow;
import net.sf.dc.core.modules.DcModule;
import net.sf.dc.core.modules.DcModules;
import net.sf.dc.core.objects.DcObject;
import net.sf.dc.core.objects.DcStorageMedium;
import net.sf.dc.messages.Messages;
import net.sf.dc.messages.MessagesWriter;
import net.sf.dc.resources.DcResources;
import net.sf.ext.util.DirectoryReader;

public abstract class ImportCDProcess extends Thread {

    protected final ImportCDDialog ui;

    protected final DcStorageMedium storageMedium;
    protected final String sourceDir;
    protected final boolean confirmDirs;
    protected final boolean recursiveDirs;
    protected final String cdName;

    protected boolean keepOnRunning = true;

    public ImportCDProcess( ImportCDDialog ui,
                            DcStorageMedium storageMedium,
                            String sourceDir,
                            boolean confirmDirs,
                            boolean recursiveDirs,
                            String cdName) {
    	this.ui = ui;
        this.storageMedium = storageMedium;

        if (!sourceDir.endsWith(File.separator)) 
            this.sourceDir = sourceDir;
        else
            this.sourceDir = sourceDir.substring(0, sourceDir.length() -1);

        this.confirmDirs = confirmDirs;
        this.recursiveDirs = recursiveDirs;
        this.cdName = cdName;
    }

    public void run() {
        try {
            ui.addMessage(DcResources.getText("msgImportStarts"));
            ui.addMessage(DcResources.getText("msgImportReadingDirs"));
            
            File file = new File(sourceDir);
            Vector filesAndDirs = new Vector();
            if (file.isFile()) {
                filesAndDirs.add(sourceDir);
            } else {
                filesAndDirs.addAll(DirectoryReader.read( 
                        sourceDir, confirmDirs, recursiveDirs, true, getExtensions(), false));
            }

            DataCrow.mainFrame.setSelectedTab(net.sf.dc.console.MainFrame._INSERTTAB);
            writeOutput(filesAndDirs);
        } catch (Exception exp) {
            ui.addMessage("msgProcessCrash");
            MessagesWriter._useLogFile = true;
            Messages.add(exp);
            new MessageBox(DcResources.getText("msgProcessCrash"), MessageBox._ERROR);
        }
    }

    public abstract String[] getExtensions();

    protected int getFileSize(String path) {
        File file = new File(path);

        long fileSize = 0;
        if (file.isDirectory()) {
            try {
                Collection files = DirectoryReader.read(path, false, true, false, null, false);

                for (Iterator iter = files.iterator(); iter.hasNext(); ) {
                    File folderFile = new File(iter.next().toString());
                    fileSize += folderFile.length();
                }
            } catch (Exception exp) {
                Messages.add(exp);
            }
        } else {
            fileSize = file.length();
        }
        return Math.round(fileSize / 1024f);
    }

    private void finish() {
        ui.addMessage(DcResources.getText("msgImportCompleted"));
    }

    public void cancel() {
    	keepOnRunning = false;
        finish();
    }

    public void writeOutput(Collection filenames) {
        try {
            ui.addMessage(DcResources.getText("msgImportFoundXResults", "" + filenames.size()));
            ui.initProgressBar(filenames.size());
            int counter = 1;
            for (   Iterator iter = filenames.iterator(); iter.hasNext()
                    && keepOnRunning; counter++) {

                String sFilename = iter.next().toString();
                ui.addMessage(DcResources.getText("msgProcessingFileX", sFilename));
                DcObject dco = parseInformation(sFilename);

                if (dco != null) {
                    dco.applyTemplate();
                    DcModule module = DcModules.get(getModule());
                    module.getCurrentInsertView().add(dco, false);
                }

                try {
                	sleep(200);
                } catch (Exception ignore) {}

                ui.updateProgressBar(counter);
            }
        } catch (Exception exp) {
            new MessageBox(ui, DcResources.getText("msgProcessCrash"),
                           MessageBox._ERROR);
            Messages.add(exp);
        } finally {
        	filenames.clear();
        	filenames = null;
            ui.addMessage(DcResources.getText("msgImportStops"));
            ui.allowActions();
        }
    }

    protected String getName(String file) {
        String name = file;
        int index = name.lastIndexOf(File.separator);
        if (index >=0) {
            name = name.substring(index + 1, name.length());

            index = name.lastIndexOf(".");
            if (index > 0 && index > name.length() - 5) {
                name = name.substring(0, index);
            }

            name = name.replaceAll("\\.", " ");
            name = name.replaceAll("_", " ");
            name = name.replaceAll("-", " ");
        }
        return name;
    }

    protected abstract DcObject parseInformation(String sFilename);

    protected abstract int getModule();
}
