/******************************************************************************
 *                                     __                                     *
 *                              <-----/@@\----->                              *
 *                             <-< <  \\//  > >->                             *
 *                               <-<-\ __ /->->                               *
 *                               Data /  \ Crow                               *
 *                                   ^    ^                                   *
 *                       (c) 2003 The Data Crow team                          *
 *                              info@datacrow.net                             *
 *                                                                            *
 *                                                                            *
 *       This library is free software; you can redistribute it and/or        *
 *        modify it under the terms of the GNU Lesser General Public          *
 *       License as published by the Free Software Foundation; either         *
 *     version 2.1 of the License, or (at your option) any later version.     *
 *                                                                            *
 *      This library is distributed in the hope that it will be useful,       *
 *      but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU       *
 *           Lesser General Public License for more details.                  *
 *                                                                            *
 *     You should have received a copy of the GNU Lesser General Public       *
 *    License along with this library; if not, write to the Free Software     *
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA   *
 *                                                                            *
 ******************************************************************************/

package net.sf.dc.processes.synchronizer;

import java.io.File;
import java.util.ArrayList;
import java.util.Collection;
import java.util.Iterator;

import net.sf.dc.console.views.View;
import net.sf.dc.core.data.DataManager;
import net.sf.dc.core.modules.DcModules;
import net.sf.dc.core.objects.DcObject;
import net.sf.dc.core.objects.musicalbum.MusicAlbum;
import net.sf.dc.core.objects.musicalbum.MusicFile;
import net.sf.dc.core.objects.musicalbum.MusicTrack;
import net.sf.dc.db.DatabaseManager;
import net.sf.dc.exceptions.ValidationException;
import net.sf.dc.resources.DcResources;
import entagged.audioformats.exceptions.CannotReadException;

public class MusicAlbumSynchronizer extends Synchronizer {

    public MusicAlbumSynchronizer() {
        super(DcResources.getText("lblMassItemUpdate", DcModules.get(DcModules._MUSICALBUM).getObjectName()),
              DcModules._MUSICALBUM);
    }
    
    public String getHelpText() {
        return DcResources.getText("msgMusicFileMassUpdateHelp");
    }
    
    public boolean canParseFiles() {
        return true;
    }

    public boolean canUseOnlineServices() {
        return true;
    }    
    
    public Thread getTask() {
        return new Task();
    }    
    
    private class Task extends Thread {
        public void run() {
            try {
                initialize();
                
                View view = DcModules.getCurrentModule().getCurrentSearchView();
                
                int items = view.getRows();
                Collection objects = new ArrayList();
                for (int i = 0; i < items && !isCancelled(); i++)
                    objects.add(view.getObject(i));
                
                initProgressBar(items);
                for (Iterator iterItems = objects.iterator(); iterItems.hasNext() && !isCancelled(); ) {
                    DcObject dco = (DcObject) iterItems.next();
                    
                    dco.loadChildren();
                    Collection children = dco.getChildren();

                    boolean updated = false;
                    
                    if (isReparseFiles()) {
                        for (Iterator iter = children.iterator(); iter.hasNext() && !isCancelled(); ) {
                            DcObject child = (DcObject) iter.next();
                            
                            String filename = child.getFilename();
                            File tst = new File(filename);
                            if (!tst.exists())
                                filename = filename.replaceAll("`", "'");
                            
                            try {
                                addMessage(DcResources.getText("msgParsing", filename));
                                
                                MusicFile musicFile = new MusicFile(filename);
        
                                dco.setValue(MusicAlbum._A_TITLE, musicFile.getAlbum());
                                dco.setValue(MusicAlbum._F_ARTIST, musicFile.getArtist());
        
                                setValue(child, MusicTrack._K_QUALITY, new Integer(musicFile.getBitrate()));
                                setValue(child, MusicTrack._J_PLAYLENGTH, new Integer(musicFile.getLength()));
                                setValue(child, MusicTrack._L_ENCODING, musicFile.getEncodingType());
                                setValue(child, MusicTrack._A_TITLE, musicFile.getTitle());
                                setValue(child, MusicTrack._G_ARTIST, musicFile.getArtist());
                                setValue(child, MusicTrack._C_YEAR, musicFile.getYear());
                                
                                
                                String sGenre = musicFile.getGenre();
                                Collection genres = DataManager.createMapping(child, sGenre, 
                                                                              DcModules._MUSICGENRE, 
                                                                              DcModules._MUSICTRACKGENREMAPPING);
                                setValue(child, MusicTrack._H_GENRES, genres);
                                setValue(child, MusicTrack._F_TRACKNUMBER, new Integer(musicFile.getTrack()));
                                
                                child.setSilence(true);
                                updated = true;
                            } catch (CannotReadException exp) {
                                child.markAsUnchanged();
                                addMessage(exp.getMessage());
                            }
                        }
                    }
                    
                    if (useOnlineService()) {
                        addMessage(DcResources.getText("msgSearchingOnlineFor", "" + dco));
                        updated = new MusicFile().onlineUpdate((MusicAlbum) dco, getServer(), getRegion());
                    }

                    updateProgressBar();
                    
                    try {
                        if (updated) {
                            dco.setSilence(true);
                            dco.saveUpdate(true);
                            
                            while (DatabaseManager.getQueueSize() > 0) {
                                try {
                                    sleep(100);
                                } catch (Exception exp) {}
                            }
                        }
                    } catch (ValidationException ve) {
                        addMessage(ve.getMessage());
                    }
                }
            } finally {
                addMessage(DcResources.getText("msgSynchronizerEnded"));
                enableAction(true);
            }  
        }
    }
}
