/******************************************************************************
 *                                     __                                     *
 *                              <-----/@@\----->                              *
 *                             <-< <  \\//  > >->                             *
 *                               <-<-\ __ /->->                               *
 *                               Data /  \ Crow                               *
 *                                   ^    ^                                   *
 *                       (c) 2003 The Data Crow team                          *
 *                              info@datacrow.net                             *
 *                                                                            *
 *                                                                            *
 *       This library is free software; you can redistribute it and/or        *
 *        modify it under the terms of the GNU Lesser General Public          *
 *       License as published by the Free Software Foundation; either         *
 *     version 2.1 of the License, or (at your option) any later version.     *
 *                                                                            *
 *      This library is distributed in the hope that it will be useful,       *
 *      but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU       *
 *           Lesser General Public License for more details.                  *
 *                                                                            *
 *     You should have received a copy of the GNU Lesser General Public       *
 *    License along with this library; if not, write to the Free Software     *
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA   *
 *                                                                            *
 ******************************************************************************/

package net.sf.dc.reporting.writer;

import java.io.IOException;
import java.util.ArrayList;
import java.util.Collection;
import java.util.Iterator;

import net.sf.dc.core.Repository;
import net.sf.dc.core.modules.DcMappingModule;
import net.sf.dc.core.modules.DcMediaModule;
import net.sf.dc.core.modules.DcModule;
import net.sf.dc.core.modules.DcModules;
import net.sf.dc.core.modules.template.ITemplateModule;
import net.sf.dc.core.objects.DcField;
import net.sf.dc.core.objects.DcMapping;
import net.sf.dc.core.objects.DcMediaObject;
import net.sf.dc.core.objects.DcObject;
import net.sf.dc.core.objects.DcSimpleObject;
import net.sf.dc.reporting.reports.XmlReference;

public class XmlSchemaWriter extends XmlBaseWriter {
    
    private Collection references = new ArrayList();
    
    public XmlSchemaWriter(String filename) throws IOException {
        super(filename);
    }
    
    public void create(DcObject dco) throws IOException {
        startDocument(dco);
        
        Collection handled = new ArrayList();
        if (dco.getModule().getIndex() == DcModules._GENERIC) {
            for (Iterator iter = DcModules.getModules(); iter.hasNext(); ) {
                DcModule module = (DcModule) iter.next();
                if (module instanceof DcMediaModule && 
                    module.isTopModule() &&
                    module.getIndex() != DcModules._GENERIC &&
                  !(module instanceof ITemplateModule)) {
                    
                    handle(module.getDcObject(), handled);
                }
            }
        } else {
            handle(dco, handled);
        }
        
        endDocument();
    }
    
    private void handle(DcObject dco, Collection handled) throws IOException {
        for (Iterator iter = dco.getFields().iterator(); iter.hasNext(); ) {
            DcField field = (DcField) iter.next();
            
            if (field.getValueType() == Repository.datatypes._DCOBJECTCOLLECTION) {
                DcMappingModule mappingMod = (DcMappingModule) DcModules.get(field.getModRef());
                DcMapping mapping = (DcMapping) mappingMod.getDcObject();
                DcModule sm = DcModules.get(mapping.getSimpleModuleIdx());
                DcObject so = sm.getDcObject();

                if (!handled.contains(so.getModule().getObjectName())) {
                    writeDco(so);
                    newLine();
                }
                
                handled.add(so.getModule().getObjectName());
            }
        }
        
        if (dco.getModule().getChildModule() != null) {
            writeDco(dco.getModule().getChildModule().getDcObject());
            newLine();
            
            handled.add(dco.getModule().getChildModule().getObjectName());
        }
        
        writeDco(dco);
        handled.add(dco.getModule().getObjectName());
    }
    
    private void addReference(String name, String reference) {
        XmlReference xmlReference = new XmlReference(name, reference);
        if (!references.contains(xmlReference))
            references.add(xmlReference);
    }
    
    private void writeField(DcField field, Collection references) throws IOException {
        String label = getValidTag(field.getSystemName());

        if (field.getValueType() == Repository.datatypes._DCOBJECTCOLLECTION) {
            DcMappingModule mappingMod = (DcMappingModule) DcModules.get(field.getModRef());
            DcMapping mapping = (DcMapping) mappingMod.getDcObject();
            DcModule sm = DcModules.get(mapping.getSimpleModuleIdx());

            String name = getValidTag(field.getSystemName());
            String reference = getValidTag(sm.getObjectName());

            writeLine("<xsd:element name=\"" + name + "\"/>", 3);
            addReference(name, reference);
        } else {
            String type;
            switch (field.getValueType()) {
            case Repository.datatypes._BIGINTEGER :
                type = "long";
                break;
            case Repository.datatypes._BOOLEAN :
                type = "boolean";
                break;
            case Repository.datatypes._DATE :
                type = "date";
                break;
            case Repository.datatypes._INTEGER :
                type = "integer";
                break;
            default:
                type = "string";
            }
            
            writeLine("<xsd:element name=\"" + label + "\" type=\"xsd:" + type + "\"/>", 3);
        }        
    }
    
    private void writeDco(DcObject dco) throws IOException {
        String baseName = getValidTag(dco.getModule().getObjectName());
        
        newLine();
        writeLine("<xsd:element name=\"" + baseName + "\" type=\"type-" + baseName + "\"/>", 1);
        writeLine("<xsd:complexType name=\"type-" + baseName + "\">", 1);
        writeLine("<xsd:sequence>", 2);
        
        if (dco instanceof DcSimpleObject) {
            String label = getValidTag(dco.getField(DcSimpleObject._A_NAME).getSystemName());
            writeLine("<xsd:element name=\"" + label + "\" type=\"xsd:string\"/>", 3);
        } else if (dco instanceof DcMediaObject) {
            
            writeField(dco.getField(DcObject._SYS_MODULE), references);
            for (Iterator iter = dco.getFields().iterator(); iter.hasNext(); ) {
                DcField field = (DcField) iter.next();
                writeField(field, references);
            }
        }
        
        if (dco.getModule().getChildModule() != null) {
            String name = getValidTag(dco.getModule().getChildModule().getObjectNamePlural());
            String reference = getValidTag(dco.getModule().getChildModule().getObjectName());
            
            writeLine("<xsd:element name=\"" + name + "\"/>", 3);
            addReference(name, reference);
        }
        
        writeLine("</xsd:sequence>", 2);
        writeLine("</xsd:complexType>", 1);
    }
    
    private void startDocument(DcObject dco) throws IOException{
        writeLine("<?xml version=\"1.0\"?>", 0);
        writeLine("<xsd:schema xmlns:xsd=\"http://www.w3.org/2001/XMLSchema\" elementFormDefault=\"qualified\">", 0);
        writeLine("<xsd:element name=\"data-crow-objects\">", 1);
        writeLine("<xsd:complexType>", 1);
        writeLine("<xsd:sequence>", 2);
        writeLine("<xsd:element maxOccurs=\"unbounded\" ref=\"" + 
                    getValidTag(dco.getModule().getObjectName())  + 
                  "\"/>", 3);
        writeLine("</xsd:sequence>", 2);
        writeLine("</xsd:complexType>", 1);
        writeLine("</xsd:element>", 0);
    }
    
    private void endDocument() throws IOException {
        writeReferences();
        writeLine("</xsd:schema>", 0);
        bos.flush();
        bos.close();
    }
    
    private void writeReferences() throws IOException {
        for (Iterator iter = references.iterator(); iter.hasNext(); ) {
            newLine();

            XmlReference reference = (XmlReference) iter.next();
            
            writeLine("<xsd:element name=\"" + reference.getName() + "\">", 1);
            writeLine("<xsd:complexType>", 2);
            writeLine("<xsd:sequence>", 3);
            writeLine("<xsd:element maxOccurs=\"unbounded\" ref=\"" + reference.getReference() + "\"/>", 4);
            writeLine("</xsd:sequence>", 3);
            writeLine("</xsd:complexType>", 2);
            writeLine("</xsd:element>", 1);
        }        
    }
}