/******************************************************************************
 *                                     __                                     *
 *                              <-----/@@\----->                              *
 *                             <-< <  \\//  > >->                             *
 *                               <-<-\ __ /->->                               *
 *                               Data /  \ Crow                               *
 *                                   ^    ^                                   *
 *                       (c) 2003 The Data Crow team                          *
 *                              info@datacrow.net                             *
 *                                                                            *
 *                                                                            *
 *       This library is free software; you can redistribute it and/or        *
 *        modify it under the terms of the GNU Lesser General Public          *
 *       License as published by the Free Software Foundation; either         *
 *     version 2.1 of the License, or (at your option) any later version.     *
 *                                                                            *
 *      This library is distributed in the hope that it will be useful,       *
 *      but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU       *
 *           Lesser General Public License for more details.                  *
 *                                                                            *
 *     You should have received a copy of the GNU Lesser General Public       *
 *    License along with this library; if not, write to the Free Software     *
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA   *
 *                                                                            *
 ******************************************************************************/

package net.sf.dc.reporting.writer;

import java.awt.Image;
import java.awt.image.BufferedImage;
import java.io.BufferedOutputStream;
import java.io.File;
import java.io.FileOutputStream;
import java.io.IOException;
import java.util.Collection;
import java.util.Iterator;

import javax.imageio.ImageIO;
import javax.swing.ImageIcon;

import net.sf.dc.core.DataCrow;
import net.sf.dc.core.modules.DcModule;
import net.sf.dc.core.objects.DcMapping;
import net.sf.dc.core.objects.DcObject;
import net.sf.dc.core.objects.DcSimpleObject;
import net.sf.dc.core.objects.Picture;
import net.sf.dc.messages.Messages;
import net.sf.dc.reporting.templates.ReportTemplateProperties;
import net.sf.ext.util.Utilities;

public class XmlWriter extends XmlBaseWriter {
    
    private int tagIdent;
    private int valueIdent;
    private ReportTemplateProperties properties;
    
    private String schemaFile;
    private final int stepSize = 4;
    
    private final String reportName;
    private final String baseDir;
    
    public XmlWriter(String filename, String schemaFile, ReportTemplateProperties properties) throws IOException {
        this(new BufferedOutputStream(new FileOutputStream(filename)), filename, schemaFile, properties);
    }
    
    public XmlWriter(BufferedOutputStream bos, String filename, String schemaFile, ReportTemplateProperties properties) {
        super(bos);
        
        this.baseDir = filename.substring(0, filename.lastIndexOf(File.separator) + 1);
        this.reportName = filename.substring(filename.lastIndexOf(File.separator) + 1, filename.lastIndexOf("."));
        this.schemaFile = schemaFile;
        this.properties = properties;
        
        resetIdent();
        
        if (properties.getBoolean(ReportTemplateProperties._COPY_IMAGES))
            createImageDir();
    }    
    
    private String getImageDir() {
        return baseDir + reportName +  "_images/";
    }
    
    private void createImageDir() {
        new File(getImageDir()).mkdirs();
    }
    
    public void resetIdent() {
        tagIdent = stepSize * 1;
        valueIdent =  stepSize * 2;
    }
    
    public void setIdent(int times) {
        tagIdent =  (stepSize * (1 * times)) + stepSize;
        valueIdent = stepSize * (2 * times);
    }
    
    public void startDocument() throws IOException {
        writeTag("<?xml version=\"1.0\"?>");
        newLine();
        writeTag("<" + uberTag + " xmlns:xsi=\"http://www.w3.org/2001/XMLSchema-instance\" xsi:noNamespaceSchemaLocation=\"file://" + schemaFile + "\">");
        newLine();
    }
    
    public void endDocument() throws IOException  {
        writeTag("</" + uberTag + ">");
        newLine();
        bos.close();
    }    
    
    public void startEntity(DcObject dco) throws IOException {
        ident(tagIdent);
        writeTag("<" + getValidTag(dco.getModule().getObjectName()) + ">");
        newLine();
    }

    public void endEntity(DcObject dco) throws IOException {
        ident(tagIdent);
        writeTag("</" + getValidTag(dco.getModule().getObjectName()) + ">");
        newLine();
    }
    
    public void writeAttribute(DcObject dco, int field) throws IOException {
        ident(valueIdent);
        String tag = getValidTag(dco.getField(field).getSystemName());
        writeTag("<" + tag + ">");
        writeValue(dco, field);
        writeTag("</" + tag + ">");
        newLine();
    }

    public void startRelations(DcModule childModule) throws IOException {
        ident(valueIdent);
        writeTag("<" + getValidTag(childModule.getObjectNamePlural()) + ">");
        newLine();
    }

    public void endRelations(DcModule childModule) throws IOException {
        ident(valueIdent);
        writeTag("</" + getValidTag(childModule.getObjectNamePlural()) + ">");
        newLine();
    }
    
    private void writeValue(DcObject dco, int field) throws IOException {
        Object o = dco.getValue(field);
        
        if (o instanceof Collection) {
            newLine();

            tagIdent += (stepSize * 2);
            valueIdent += (stepSize * 2);
            for (Iterator iter = ((Collection) o).iterator(); iter.hasNext(); ){
                DcObject subDco = (DcObject) iter.next();
                
                if (subDco instanceof DcMapping)
                    subDco = ((DcMapping) subDco).getSimpleObject();
                
                startEntity(subDco);
                writeAttribute(subDco, DcSimpleObject._A_NAME);
                endEntity(subDco);
            }
            valueIdent -= (stepSize * 2);
            tagIdent -= (stepSize * 2);
            ident(valueIdent);            
            
        } else if (o instanceof Picture) {
            Picture picture = (Picture) o;

            String filename = (String) picture.getValue(Picture._C_FILENAME);
            if (properties.getBoolean(ReportTemplateProperties._COPY_IMAGES)) {
                copyImage((Picture) o, getImageDir() + filename);
                filename = "./" + reportName +  "_images/" + filename;
            } else {
                filename = "file://" + DataCrow.imageDir + picture.getValue(Picture._C_FILENAME);
            }
            
            write(filename);
            
        } else {
            String text = dco.getDisplayString(field);
            
            int maximumLength = properties.getInt(ReportTemplateProperties._MAX_TEXT_LENGTH);
            if (maximumLength > 0 && text.length() > maximumLength) {
                text = text.substring(0, maximumLength);
                text = text.substring(0, text.lastIndexOf(" ")) + "...";
            }
            
            write(text);
        }
    }
    
    private void copyImage(Picture picture, String target) {
        
        try {
            picture.loadImage();
            byte[] image = (byte[]) picture.getValue(Picture._D_IMAGE);

            if (image != null) {
                
                if (properties.getBoolean(ReportTemplateProperties._SCALE_IMAGES)) {
                    int width = properties.getInt(ReportTemplateProperties._IMAGE_WIDTH);
                    int height = properties.getInt(ReportTemplateProperties._IMAGE_HEIGHT);

                    Image img = Utilities.getScaledImage(image, width, height);
                    BufferedImage bufferedImage = Utilities.toBufferedImage(new ImageIcon(img));
                    ImageIO.write(bufferedImage, "jpg", new File(target));
                    bufferedImage.flush();
                    img.flush();
                } else {
                    Utilities.writeImageToFile(image, target);
                }
            }
            
        } catch (Exception e) {
            Messages.add(e);
        }
    }
    
    private void ident(int x) throws IOException {
        String s = "";
        for (int i = 0; i < x; i++)
            s += " ";
        bos.write(s.getBytes());
    }    
    
    private void write(String value) throws IOException {
        value = value.replaceAll("&", "&amp;");
        value = value.replaceAll("<", "&lt;");
        value = value.replaceAll(">", "&gt;");
        value = value.replaceAll("\"", "&quot;");
        value = value.replaceAll("'", "&apos;");
        bos.write(value.getBytes("UTF8"));
    }    
}
